import os
import tempfile

import numpy as np
import pytest

from matio import load_from_mat, save_to_mat
from matio.utils.matclass import MatlabOpaque, MatlabOpaqueArray

files = [("test_user_defined_v7.mat", "v7"), ("test_user_defined_v73.mat", "v7.3")]
namespace = "TestClasses"


@pytest.mark.parametrize("filename, version", files)
class TestLoadMatlabUserDefined:

    def test_obj_no_vals(self, filename, version):
        """Test reading uninitialized object from MAT-file"""
        file_path = os.path.join(os.path.dirname(__file__), "data", filename)
        mdict = load_from_mat(file_path, variable_names=["obj_no_vals"])
        assert "obj_no_vals" in mdict
        assert isinstance(mdict["obj_no_vals"], MatlabOpaque)
        assert mdict["obj_no_vals"].classname == f"{namespace}.BasicClass"
        assert mdict["obj_no_vals"].type_system == "MCOS"

        prop_dict = {
            "a": np.empty((0, 0), dtype=np.float64),
            "b": np.empty((0, 0), dtype=np.float64),
            "c": np.empty((0, 0), dtype=np.float64),
        }

        for key, val in prop_dict.items():
            assert key in mdict["obj_no_vals"].properties
            np.testing.assert_array_equal(
                mdict["obj_no_vals"].properties[key], val, strict=True
            )

    def test_obj_with_vals(self, filename, version):
        """Test reading initialized object from MAT-file"""
        file_path = os.path.join(os.path.dirname(__file__), "data", filename)
        mdict = load_from_mat(file_path, variable_names=["obj_with_vals"])
        assert "obj_with_vals" in mdict
        assert isinstance(mdict["obj_with_vals"], MatlabOpaque)
        assert mdict["obj_with_vals"].classname == f"{namespace}.BasicClass"
        assert mdict["obj_with_vals"].type_system == "MCOS"

        prop_dict = {
            "a": np.array([[10]], dtype=np.float64),
            "b": np.empty((0, 0), dtype=np.float64),
            "c": np.empty((0, 0), dtype=np.float64),
        }

        for key, val in prop_dict.items():
            assert key in mdict["obj_with_vals"].properties
            np.testing.assert_array_equal(
                mdict["obj_with_vals"].properties[key], val, strict=True
            )

    def test_obj_with_default_val(self, filename, version):
        """Test reading object with default values (including objects in default vals) from MAT-file"""
        file_path = os.path.join(os.path.dirname(__file__), "data", filename)
        mdict = load_from_mat(file_path, variable_names=["obj_with_default_val"])
        assert "obj_with_default_val" in mdict
        assert isinstance(mdict["obj_with_default_val"], MatlabOpaque)
        assert mdict["obj_with_default_val"].classname == f"{namespace}.DefaultClass"
        assert mdict["obj_with_default_val"].type_system == "MCOS"

        prop_dict = {
            "a": np.array(
                [["Default String"]], dtype=np.dtypes.StringDType(na_object=np.nan)
            ),
            "b": np.array([[10]], dtype=np.float64),
        }

        assert "a" in mdict["obj_with_default_val"].properties
        np.testing.assert_array_equal(
            mdict["obj_with_default_val"].properties["a"], prop_dict["a"], strict=True
        )

        assert "b" in mdict["obj_with_default_val"].properties
        np.testing.assert_array_equal(
            mdict["obj_with_default_val"].properties["b"], prop_dict["b"], strict=True
        )

    def test_obj_array(self, filename, version):
        """Test reading object array from MAT-file"""
        file_path = os.path.join(os.path.dirname(__file__), "data", filename)
        mdict = load_from_mat(file_path, variable_names=["obj_array"])
        assert "obj_array" in mdict
        assert isinstance(mdict["obj_array"], MatlabOpaqueArray)
        assert mdict["obj_array"].classname == f"{namespace}.BasicClass"
        assert mdict["obj_array"].type_system == "MCOS"
        assert mdict["obj_array"].shape == (2, 2)
        assert mdict["obj_array"].dtype == object
        assert all(isinstance(x, MatlabOpaque) for x in mdict["obj_array"].flat)

        np.testing.assert_array_equal(
            mdict["obj_array"][0, 0].properties["a"],
            np.array([[1]], dtype=np.float64),
            strict=True,
        )
        np.testing.assert_array_equal(
            mdict["obj_array"][0, 1].properties["a"],
            np.array([[2]], dtype=np.float64),
            strict=True,
        )
        np.testing.assert_array_equal(
            mdict["obj_array"][1, 0].properties["a"],
            np.array([[3]], dtype=np.float64),
            strict=True,
        )
        np.testing.assert_array_equal(
            mdict["obj_array"][1, 1].properties["a"],
            np.array([[4]], dtype=np.float64),
            strict=True,
        )

    def test_obj_with_nested_props(self, filename, version):
        """Test reading objects with nested properties from MAT-file"""
        file_path = os.path.join(os.path.dirname(__file__), "data", filename)
        mdict = load_from_mat(file_path, variable_names=["obj_with_nested_props"])
        assert "obj_with_nested_props" in mdict
        assert isinstance(mdict["obj_with_nested_props"], MatlabOpaque)
        assert mdict["obj_with_nested_props"].classname == f"{namespace}.BasicClass"
        assert mdict["obj_with_nested_props"].type_system == "MCOS"

        assert "a" in mdict["obj_with_nested_props"].properties
        assert isinstance(mdict["obj_with_nested_props"].properties["a"], MatlabOpaque)
        assert (
            mdict["obj_with_nested_props"].properties["a"].classname
            == f"{namespace}.BasicClass"
        )
        prop_a = {
            "a": np.array([[1]], dtype=np.float64),
            "b": np.array(["Obj1"], dtype=np.str_),
            "c": np.empty((0, 0), dtype=np.float64),
        }
        for key, val in prop_a.items():
            assert key in mdict["obj_with_nested_props"].properties["a"].properties
            np.testing.assert_array_equal(
                mdict["obj_with_nested_props"].properties["a"].properties[key],
                val,
                strict=True,
            )

        assert "b" in mdict["obj_with_nested_props"].properties
        assert isinstance(mdict["obj_with_nested_props"].properties["b"], np.ndarray)
        assert mdict["obj_with_nested_props"].properties["b"].dtype == object
        assert mdict["obj_with_nested_props"].properties["b"].shape == (1, 1)
        assert isinstance(
            mdict["obj_with_nested_props"].properties["b"][0, 0], MatlabOpaque
        )
        assert (
            mdict["obj_with_nested_props"].properties["b"][0, 0].classname
            == f"{namespace}.BasicClass"
        )
        for key, val in prop_a.items():
            assert (
                key in mdict["obj_with_nested_props"].properties["b"][0, 0].properties
            )
            np.testing.assert_array_equal(
                mdict["obj_with_nested_props"].properties["b"][0, 0].properties[key],
                val,
                strict=True,
            )

        assert "c" in mdict["obj_with_nested_props"].properties
        assert isinstance(mdict["obj_with_nested_props"].properties["c"], np.ndarray)
        assert set(mdict["obj_with_nested_props"].properties["c"].dtype.names) == {
            "InnerProp"
        }
        assert mdict["obj_with_nested_props"].properties["c"].shape == (1, 1)
        assert isinstance(
            mdict["obj_with_nested_props"].properties["c"][0, 0]["InnerProp"],
            MatlabOpaque,
        )
        assert (
            mdict["obj_with_nested_props"].properties["c"][0, 0]["InnerProp"].classname
            == f"{namespace}.BasicClass"
        )
        prop_c = {
            "a": np.array([[2]], dtype=np.float64),
            "b": np.array(["Obj2"], dtype=np.str_),
            "c": np.empty((0, 0), dtype=np.float64),
        }
        for key, val in prop_c.items():
            assert (
                key
                in mdict["obj_with_nested_props"]
                .properties["c"][0, 0]["InnerProp"]
                .properties
            )
            np.testing.assert_array_equal(
                mdict["obj_with_nested_props"]
                .properties["c"][0, 0]["InnerProp"]
                .properties[key],
                val,
                strict=True,
            )

    def test_handle_class_obj(self, filename, version):
        """Test reading handle class objects from MAT-file"""
        file_path = os.path.join(os.path.dirname(__file__), "data", filename)
        mdict = load_from_mat(
            file_path, variable_names=["obj_handle_1", "obj_handle_2"]
        )
        assert "obj_handle_1" in mdict
        assert "obj_handle_2" in mdict
        assert isinstance(mdict["obj_handle_1"], MatlabOpaque)
        assert isinstance(mdict["obj_handle_2"], MatlabOpaque)
        assert mdict["obj_handle_1"] is mdict["obj_handle_2"]

        assert mdict["obj_handle_1"].classname == f"{namespace}.HandleClass"
        assert mdict["obj_handle_1"].type_system == "MCOS"

        val = np.array([[20]], dtype=np.float64)
        np.testing.assert_array_equal(
            mdict["obj_handle_1"].properties["a"], val, strict=True
        )


@pytest.mark.parametrize("filename, version", files)
class TestSaveMatlabUserDefined:

    def test_obj_no_vals(self, filename, version):
        """Test writing uninitialized objects to MAT-file"""
        file_path = os.path.join(os.path.dirname(__file__), "data", filename)
        mdict = load_from_mat(file_path, variable_names=["obj_no_vals"])

        with tempfile.NamedTemporaryFile(suffix=".mat", delete=False) as tmpfile:
            temp_file_path = tmpfile.name

        try:
            save_to_mat(temp_file_path, mdict, version=version)
            mload = load_from_mat(temp_file_path, variable_names=["obj_no_vals"])

            assert isinstance(mload["obj_no_vals"], MatlabOpaque)
            assert mload["obj_no_vals"].classname == f"{namespace}.BasicClass"
            assert mload["obj_no_vals"].type_system == "MCOS"

            for key, val in mdict["obj_no_vals"].properties.items():
                assert key in mload["obj_no_vals"].properties
                np.testing.assert_array_equal(
                    mload["obj_no_vals"].properties[key], val, strict=True
                )

        finally:
            if os.path.exists(temp_file_path):
                os.remove(temp_file_path)

    def test_obj_with_vals(self, filename, version):
        """Test writing initialized objects to MAT-file"""
        file_path = os.path.join(os.path.dirname(__file__), "data", filename)
        mdict = load_from_mat(file_path, variable_names=["obj_with_vals"])

        with tempfile.NamedTemporaryFile(suffix=".mat", delete=False) as tmpfile:
            temp_file_path = tmpfile.name

        try:
            save_to_mat(temp_file_path, mdict, version=version)
            mload = load_from_mat(temp_file_path, variable_names=["obj_with_vals"])

            assert isinstance(mload["obj_with_vals"], MatlabOpaque)
            assert mload["obj_with_vals"].classname == f"{namespace}.BasicClass"
            assert mload["obj_with_vals"].type_system == "MCOS"

            for key, val in mdict["obj_with_vals"].properties.items():
                assert key in mload["obj_with_vals"].properties
                np.testing.assert_array_equal(
                    mload["obj_with_vals"].properties[key], val, strict=True
                )

        finally:
            if os.path.exists(temp_file_path):
                os.remove(temp_file_path)

    def test_obj_with_default_val(self, filename, version):
        """Test writing objects with default values (including objects in default vals) to MAT-file"""
        file_path = os.path.join(os.path.dirname(__file__), "data", filename)
        mdict = load_from_mat(file_path, variable_names=["obj_with_default_val"])

        with tempfile.NamedTemporaryFile(suffix=".mat", delete=False) as tmpfile:
            temp_file_path = tmpfile.name

        try:
            save_to_mat(temp_file_path, mdict, version=version)
            mload = load_from_mat(
                temp_file_path, variable_names=["obj_with_default_val"]
            )

            assert isinstance(mload["obj_with_default_val"], MatlabOpaque)
            assert (
                mload["obj_with_default_val"].classname == f"{namespace}.DefaultClass"
            )
            assert mload["obj_with_default_val"].type_system == "MCOS"

            for key, val in mdict["obj_with_default_val"].properties.items():
                assert key in mload["obj_with_default_val"].properties
                np.testing.assert_array_equal(
                    mload["obj_with_default_val"].properties[key], val, strict=True
                )

        finally:
            if os.path.exists(temp_file_path):
                os.remove(temp_file_path)

    def test_obj_array(self, filename, version):
        """Test writing object arrays to MAT-file"""
        file_path = os.path.join(os.path.dirname(__file__), "data", filename)
        mdict = load_from_mat(file_path, variable_names=["obj_array"])

        with tempfile.NamedTemporaryFile(suffix=".mat", delete=False) as tmpfile:
            temp_file_path = tmpfile.name

        try:
            save_to_mat(temp_file_path, mdict, version=version)
            mload = load_from_mat(temp_file_path, variable_names=["obj_array"])

            assert isinstance(mload["obj_array"], MatlabOpaqueArray)
            assert mload["obj_array"].classname == f"{namespace}.BasicClass"
            assert mload["obj_array"].type_system == "MCOS"
            assert mload["obj_array"].shape == mdict["obj_array"].shape
            assert mload["obj_array"].dtype == object
            assert all(isinstance(x, MatlabOpaque) for x in mload["obj_array"].flat)

            for idx in np.ndindex(mdict["obj_array"].shape):
                for key, val in mdict["obj_array"][idx].properties.items():
                    assert key in mload["obj_array"][idx].properties
                    np.testing.assert_array_equal(
                        mload["obj_array"][idx].properties[key], val, strict=True
                    )

        finally:
            if os.path.exists(temp_file_path):
                os.remove(temp_file_path)

    def test_obj_with_nested_props(self, filename, version):
        """Test writing objects with nested properties to MAT-file"""
        file_path = os.path.join(os.path.dirname(__file__), "data", filename)
        mdict = load_from_mat(file_path, variable_names=["obj_with_nested_props"])

        with tempfile.NamedTemporaryFile(suffix=".mat", delete=False) as tmpfile:
            temp_file_path = tmpfile.name

        try:
            save_to_mat(temp_file_path, mdict, version=version)
            mload = load_from_mat(
                temp_file_path, variable_names=["obj_with_nested_props"]
            )

            assert isinstance(mload["obj_with_nested_props"], MatlabOpaque)
            assert mload["obj_with_nested_props"].classname == f"{namespace}.BasicClass"
            assert mload["obj_with_nested_props"].type_system == "MCOS"

            assert isinstance(
                mload["obj_with_nested_props"].properties["a"], MatlabOpaque
            )
            for key, val in (
                mdict["obj_with_nested_props"].properties["a"].properties.items()
            ):
                assert key in mload["obj_with_nested_props"].properties["a"].properties
                np.testing.assert_array_equal(
                    mload["obj_with_nested_props"].properties["a"].properties[key],
                    val,
                    strict=True,
                )

            assert isinstance(
                mdict["obj_with_nested_props"].properties["b"][0, 0], MatlabOpaque
            )

            for key, val in (
                mdict["obj_with_nested_props"].properties["b"][0, 0].properties.items()
            ):
                assert (
                    key
                    in mload["obj_with_nested_props"].properties["b"][0, 0].properties
                )
                np.testing.assert_array_equal(
                    mload["obj_with_nested_props"]
                    .properties["b"][0, 0]
                    .properties[key],
                    val,
                    strict=True,
                )

            for key, val in (
                mdict["obj_with_nested_props"]
                .properties["c"][0, 0]["InnerProp"]
                .properties.items()
            ):
                assert (
                    key
                    in mload["obj_with_nested_props"]
                    .properties["c"][0, 0]["InnerProp"]
                    .properties
                )
                np.testing.assert_array_equal(
                    mload["obj_with_nested_props"]
                    .properties["c"][0, 0]["InnerProp"]
                    .properties[key],
                    val,
                    strict=True,
                )

        finally:
            if os.path.exists(temp_file_path):
                os.remove(temp_file_path)

    def test_handle_class_obj(self, filename, version):
        """Test writing handle class objects to MAT-file"""
        file_path = os.path.join(os.path.dirname(__file__), "data", filename)
        mdict = load_from_mat(
            file_path, variable_names=["obj_handle_1", "obj_handle_2"]
        )

        with tempfile.NamedTemporaryFile(suffix=".mat", delete=False) as tmpfile:
            temp_file_path = tmpfile.name

        try:
            save_to_mat(temp_file_path, mdict, version=version)
            mload = load_from_mat(
                temp_file_path, variable_names=["obj_handle_1", "obj_handle_2"]
            )

            assert isinstance(mload["obj_handle_1"], MatlabOpaque)
            assert isinstance(mload["obj_handle_2"], MatlabOpaque)
            assert mload["obj_handle_1"] is mload["obj_handle_2"]

            for key, val in mdict["obj_handle_1"].properties.items():
                assert key in mload["obj_handle_1"].properties
                np.testing.assert_array_equal(
                    mload["obj_handle_1"].properties[key], val, strict=True
                )

        finally:
            if os.path.exists(temp_file_path):
                os.remove(temp_file_path)



import os
import tempfile

import numpy as np
import pandas as pd
import pytest

from matio import load_from_mat, save_to_mat
from matio.utils.matclass import MatConvertWarning, MatlabOpaque

files = [("test_tables_v7.mat", "v7"), ("test_tables_v73.mat", "v7.3")]
namespace = "TestClasses"


@pytest.mark.parametrize("filename, version", files)
class TestLoadMatlabTimetable:

    def test_timetable_datetime(self, filename, version):
        """Test reading timetable with datetime index from MAT-file"""
        file_path = os.path.join(os.path.dirname(__file__), "data", filename)
        mdict = load_from_mat(file_path, variable_names=["timetable_datetime"])
        assert "timetable_datetime" in mdict

        df = pd.DataFrame(
            {"data1": [1.0, 2.0, 3.0]},
            index=pd.Index(
                np.array(
                    ["2023-01-01", "2023-01-02", "2023-01-03"],
                    dtype="datetime64[ns]",
                ),
                name="Time",
            ),
        )
        pd.testing.assert_frame_equal(mdict["timetable_datetime"], df, check_like=True)

    def test_timetable_duration(self, filename, version):
        """Test reading timetable with duration index from MAT-file"""
        file_path = os.path.join(os.path.dirname(__file__), "data", filename)
        mdict = load_from_mat(file_path, variable_names=["timetable_duration"])
        assert "timetable_duration" in mdict

        df = pd.DataFrame(
            {
                "data1": [1.0, 2.0, 3.0],
            },
            index=pd.Index(
                np.array([10, 20, 30], dtype="timedelta64[s]"),
                name="Time",
            ),
        )
        pd.testing.assert_frame_equal(mdict["timetable_duration"], df, check_like=True)

    def test_timetable_empty(self, filename, version):
        """Test reading empty timetable from MAT-file"""
        file_path = os.path.join(os.path.dirname(__file__), "data", filename)
        mdict = load_from_mat(file_path, variable_names=["timetable_empty"])
        assert "timetable_empty" in mdict

        df = pd.DataFrame(
            index=pd.DatetimeIndex([], dtype="datetime64[ns]", name="Time")
        )
        pd.testing.assert_frame_equal(mdict["timetable_empty"], df, check_like=True)

    def test_timetable_from_duration(self, filename, version):
        """Test reading timetable with duration starttime from MAT-file"""
        file_path = os.path.join(os.path.dirname(__file__), "data", filename)
        with pytest.warns(MatConvertWarning):
            mdict = load_from_mat(file_path, variable_names=["timetable_from_duration"])
        assert "timetable_from_duration" in mdict

        df = pd.DataFrame(
            {
                "data1": [1.0, 2.0, 3.0],
            },
            index=pd.Index(
                np.array([0, 1, 2], dtype="timedelta64[s]"),
                name="Time",
            ),
        )
        pd.testing.assert_frame_equal(
            mdict["timetable_from_duration"], df, check_like=True
        )

    def test_timetable_from_sample_rate(self, filename, version):
        """Test reading timetable with sample rate and timestep from MAT-file"""
        file_path = os.path.join(os.path.dirname(__file__), "data", filename)
        with pytest.warns(MatConvertWarning, match="get_row_times"):
            mdict = load_from_mat(
                file_path, variable_names=["timetable_from_sample_rate"]
            )
        assert "timetable_from_sample_rate" in mdict

        df = pd.DataFrame(
            {
                "data1": [1.0, 2.0, 3.0],
            },
            index=pd.Index(
                np.array([0, 100000, 200000], dtype="timedelta64[ns]"),
                name="Time",
            ),
        )
        pd.testing.assert_frame_equal(
            mdict["timetable_from_sample_rate"], df, check_like=True
        )

    def test_timetable_from_starttime_calendarDuration(self, filename, version):
        """Test reading timetable with calendarDuration starttime in MAT-file"""
        file_path = os.path.join(os.path.dirname(__file__), "data", filename)
        mdict = load_from_mat(
            file_path, variable_names=["timetable_from_starttime_calendarDuration"]
        )
        assert "timetable_from_starttime_calendarDuration" in mdict

        df = pd.DataFrame(
            {
                "data1": [1.0, 2.0, 3.0],
            },
            index=pd.Index(
                np.array(["2020-01", "2020-04", "2020-07"], dtype="datetime64[ns]"),
                name="Time",
            ),
        )
        pd.testing.assert_frame_equal(
            mdict["timetable_from_starttime_calendarDuration"], df, check_like=True
        )

    def test_timetable_from_starttime_datetime(self, filename, version):
        """Test reading timetable with datetime starttime in MAT-file"""
        file_path = os.path.join(os.path.dirname(__file__), "data", filename)
        mdict = load_from_mat(
            file_path, variable_names=["timetable_from_starttime_datetime"]
        )
        assert "timetable_from_starttime_datetime" in mdict

        df = pd.DataFrame(
            {
                "data1": [1.0, 2.0, 3.0],
            },
            index=pd.Index(
                np.array(
                    [
                        "2020-01-01T00:00:00",
                        "2020-01-01T00:00:01",
                        "2020-01-01T00:00:02",
                    ],
                    dtype="datetime64[ns]",
                ),
                name="Time",
            ),
        )
        pd.testing.assert_frame_equal(
            mdict["timetable_from_starttime_datetime"], df, check_like=True
        )

    def test_timetable_from_starttime_duration(self, filename, version):
        """Test reading timetable with duration starttime in MAT-file"""
        file_path = os.path.join(os.path.dirname(__file__), "data", filename)
        mdict = load_from_mat(
            file_path, variable_names=["timetable_from_starttime_duration"]
        )
        assert "timetable_from_starttime_duration" in mdict

        df = pd.DataFrame(
            {
                "data1": [1.0, 2.0, 3.0],
            },
            index=pd.Index(
                np.array([10, 11, 12], dtype="timedelta64[s]"),
                name="Time",
            ),
        )
        pd.testing.assert_frame_equal(
            mdict["timetable_from_starttime_duration"], df, check_like=True
        )

    def test_timetable_var_names(self, filename, version):
        """Test reading timetable with variable names from MAT-file"""
        file_path = os.path.join(os.path.dirname(__file__), "data", filename)
        mdict = load_from_mat(file_path, variable_names=["timetable_var_names"])
        assert "timetable_var_names" in mdict

        df = pd.DataFrame(
            {"Pressure": [1.0, 2.0, 3.0]},
            index=pd.Index(
                np.array(
                    ["2023-01-01", "2023-01-02", "2023-01-03"],
                    dtype="datetime64[ns]",
                ),
                name="Time",
            ),
        )
        pd.testing.assert_frame_equal(mdict["timetable_var_names"], df, check_like=True)

    def test_timetable_multi_col(self, filename, version):
        """Test reading timetable with multi-column data from MAT-file"""
        file_path = os.path.join(os.path.dirname(__file__), "data", filename)
        mdict = load_from_mat(file_path, variable_names=["timetable_multi_col"])
        assert "timetable_multi_col" in mdict

        df = pd.DataFrame(
            {
                "data2_1": [1.0, 2.0, 3.0],
                "data2_2": [4.0, 5.0, 6.0],
            },
            index=pd.Index(
                np.array([10, 20, 30], dtype="timedelta64[s]"),
                name="Time",
            ),
        )
        pd.testing.assert_frame_equal(mdict["timetable_multi_col"], df, check_like=True)

    def test_timetable_with_attrs(self, filename, version):
        """Test reading timetable with attributes from MAT-file"""
        file_path = os.path.join(os.path.dirname(__file__), "data", filename)
        mdict = load_from_mat(
            file_path, variable_names=["timetable_with_attrs"], add_table_attrs=True
        )
        assert "timetable_with_attrs" in mdict

        df = pd.DataFrame(
            {"data1": [1.0, 2.0, 3.0]},
            index=pd.Index(
                np.array(
                    ["2023-01-01", "2023-01-02", "2023-01-03"],
                    dtype="datetime64[ns]",
                ),
                name="Date",
            ),
        )
        pd.testing.assert_frame_equal(
            mdict["timetable_with_attrs"], df, check_like=True
        )

        attrs = {
            "Description": "Random Description",
            "varUnits": ["m/s"],
            "varDescriptions": ["myVar"],
            "varContinuity": ["continuous"],
            "UserData": np.empty((0, 0), dtype=np.float64),
        }

        for key, value in mdict["timetable_with_attrs"].attrs.items():
            assert key in attrs
            if isinstance(value, np.ndarray):
                np.testing.assert_array_equal(value, attrs[key])
            else:
                assert value == attrs[key]


@pytest.mark.parametrize("filename, version", files)
class TestSaveMatlabTimetable:

    def test_timetable_datetime(self, filename, version):
        """Test writing numeric timetable to MAT-file"""
        file_path = os.path.join(os.path.dirname(__file__), "data", filename)
        mdict = load_from_mat(file_path, variable_names=["timetable_datetime"])

        with tempfile.NamedTemporaryFile(suffix=".mat", delete=False) as tmpfile:
            temp_file_path = tmpfile.name

        try:
            save_to_mat(temp_file_path, mdict, version=version)
            mload = load_from_mat(temp_file_path, variable_names=["timetable_datetime"])

            pd.testing.assert_frame_equal(
                mdict["timetable_datetime"],
                mload["timetable_datetime"],
                check_like=True,
            )

        finally:
            if os.path.exists(temp_file_path):
                os.remove(temp_file_path)

    def test_timetable_duration(self, filename, version):
        """Test writing numeric timetable to MAT-file"""
        file_path = os.path.join(os.path.dirname(__file__), "data", filename)
        mdict = load_from_mat(file_path, variable_names=["timetable_duration"])

        with tempfile.NamedTemporaryFile(suffix=".mat", delete=False) as tmpfile:
            temp_file_path = tmpfile.name

        try:
            save_to_mat(temp_file_path, mdict, version=version)
            mload = load_from_mat(temp_file_path, variable_names=["timetable_duration"])

            pd.testing.assert_frame_equal(
                mdict["timetable_duration"],
                mload["timetable_duration"],
                check_like=True,
            )

        finally:
            if os.path.exists(temp_file_path):
                os.remove(temp_file_path)

    def test_timetable_empty(self, filename, version):
        """Test writing numeric timetable to MAT-file"""
        file_path = os.path.join(os.path.dirname(__file__), "data", filename)
        mdict = load_from_mat(file_path, variable_names=["timetable_empty"])

        with tempfile.NamedTemporaryFile(suffix=".mat", delete=False) as tmpfile:
            temp_file_path = tmpfile.name

        try:
            save_to_mat(temp_file_path, mdict, version=version)
            mload = load_from_mat(temp_file_path, variable_names=["timetable_empty"])

            pd.testing.assert_frame_equal(
                mdict["timetable_empty"], mload["timetable_empty"], check_like=True
            )

        finally:
            if os.path.exists(temp_file_path):
                os.remove(temp_file_path)

    def test_timetable_from_duration(self, filename, version):
        """Test writing numeric timetable to MAT-file"""
        file_path = os.path.join(os.path.dirname(__file__), "data", filename)
        with pytest.warns(MatConvertWarning):
            mdict = load_from_mat(file_path, variable_names=["timetable_from_duration"])

        with tempfile.NamedTemporaryFile(suffix=".mat", delete=False) as tmpfile:
            temp_file_path = tmpfile.name

        try:
            save_to_mat(temp_file_path, mdict, version=version)
            mload = load_from_mat(
                temp_file_path, variable_names=["timetable_from_duration"]
            )

            pd.testing.assert_frame_equal(
                mdict["timetable_from_duration"],
                mload["timetable_from_duration"],
                check_like=True,
            )

        finally:
            if os.path.exists(temp_file_path):
                os.remove(temp_file_path)

    def test_timetable_from_starttime_calendarDuration(self, filename, version):
        """Test writing numeric timetable to MAT-file"""
        file_path = os.path.join(os.path.dirname(__file__), "data", filename)
        mdict = load_from_mat(
            file_path, variable_names=["timetable_from_starttime_calendarDuration"]
        )

        with tempfile.NamedTemporaryFile(suffix=".mat", delete=False) as tmpfile:
            temp_file_path = tmpfile.name

        try:
            save_to_mat(temp_file_path, mdict, version=version)
            mload = load_from_mat(
                temp_file_path,
                variable_names=["timetable_from_starttime_calendarDuration"],
            )

            pd.testing.assert_frame_equal(
                mdict["timetable_from_starttime_calendarDuration"],
                mload["timetable_from_starttime_calendarDuration"],
                check_like=True,
            )

        finally:
            if os.path.exists(temp_file_path):
                os.remove(temp_file_path)

    def test_timetable_from_starttime_datetime(self, filename, version):
        """Test writing numeric timetable to MAT-file"""
        file_path = os.path.join(os.path.dirname(__file__), "data", filename)
        mdict = load_from_mat(
            file_path, variable_names=["timetable_from_starttime_datetime"]
        )

        with tempfile.NamedTemporaryFile(suffix=".mat", delete=False) as tmpfile:
            temp_file_path = tmpfile.name

        try:
            save_to_mat(temp_file_path, mdict, version=version)
            mload = load_from_mat(
                temp_file_path, variable_names=["timetable_from_starttime_datetime"]
            )

            pd.testing.assert_frame_equal(
                mdict["timetable_from_starttime_datetime"],
                mload["timetable_from_starttime_datetime"],
                check_like=True,
            )

        finally:
            if os.path.exists(temp_file_path):
                os.remove(temp_file_path)

    def test_timetable_from_starttime_duration(self, filename, version):
        """Test writing numeric timetable to MAT-file"""
        file_path = os.path.join(os.path.dirname(__file__), "data", filename)
        mdict = load_from_mat(
            file_path, variable_names=["timetable_from_starttime_duration"]
        )

        with tempfile.NamedTemporaryFile(suffix=".mat", delete=False) as tmpfile:
            temp_file_path = tmpfile.name

        try:
            save_to_mat(temp_file_path, mdict, version=version)
            mload = load_from_mat(
                temp_file_path, variable_names=["timetable_from_starttime_duration"]
            )

            pd.testing.assert_frame_equal(
                mdict["timetable_from_starttime_duration"],
                mload["timetable_from_starttime_duration"],
                check_like=True,
            )

        finally:
            if os.path.exists(temp_file_path):
                os.remove(temp_file_path)

    def test_timetable_var_names(self, filename, version):
        """Test writing numeric timetable to MAT-file"""
        file_path = os.path.join(os.path.dirname(__file__), "data", filename)
        mdict = load_from_mat(file_path, variable_names=["timetable_var_names"])

        with tempfile.NamedTemporaryFile(suffix=".mat", delete=False) as tmpfile:
            temp_file_path = tmpfile.name

        try:
            save_to_mat(temp_file_path, mdict, version=version)
            mload = load_from_mat(
                temp_file_path, variable_names=["timetable_var_names"]
            )

            pd.testing.assert_frame_equal(
                mdict["timetable_var_names"],
                mload["timetable_var_names"],
                check_like=True,
            )

        finally:
            if os.path.exists(temp_file_path):
                os.remove(temp_file_path)

    def test_timetable_multi_col(self, filename, version):
        """Test writing numeric timetable to MAT-file"""
        file_path = os.path.join(os.path.dirname(__file__), "data", filename)
        mdict = load_from_mat(file_path, variable_names=["timetable_multi_col"])

        with tempfile.NamedTemporaryFile(suffix=".mat", delete=False) as tmpfile:
            temp_file_path = tmpfile.name

        try:
            save_to_mat(temp_file_path, mdict, version=version)
            mload = load_from_mat(
                temp_file_path, variable_names=["timetable_multi_col"]
            )

            pd.testing.assert_frame_equal(
                mdict["timetable_multi_col"],
                mload["timetable_multi_col"],
                check_like=True,
            )

        finally:
            if os.path.exists(temp_file_path):
                os.remove(temp_file_path)

    @pytest.mark.xfail(reason="Writing attributes not implemented")
    def test_timetable_with_attrs(self, filename, version):
        """Test writing numeric timetable to MAT-file"""
        file_path = os.path.join(os.path.dirname(__file__), "data", filename)
        mdict = load_from_mat(
            file_path, variable_names=["timetable_with_attrs"], add_table_attrs=True
        )

        with tempfile.NamedTemporaryFile(suffix=".mat", delete=False) as tmpfile:
            temp_file_path = tmpfile.name

        try:
            save_to_mat(temp_file_path, mdict, version=version)
            mload = load_from_mat(
                temp_file_path,
                variable_names=["timetable_with_attrs"],
                add_table_attrs=True,
            )

            pd.testing.assert_frame_equal(
                mdict["timetable_with_attrs"],
                mload["timetable_with_attrs"],
                check_like=True,
            )

            for key, value in mdict["timetable_with_attrs"].attrs.items():
                assert key in mload["timetable_with_attrs"].attrs
                if isinstance(value, np.ndarray):
                    np.testing.assert_array_equal(
                        value, mload["timetable_with_attrs"].attrs[key]
                    )
                else:
                    assert value == mload["timetable_with_attrs"].attrs[key]

        finally:
            if os.path.exists(temp_file_path):
                os.remove(temp_file_path)



"""Utility functions for converting MATLAB containerMap and Dictionary"""

import warnings

import numpy as np

from matio.utils.matclass import MatlabContainerMap

MAT_DICT_VERSION = 1

map_dtype_map = {
    np.dtype("float64"): "double",
    np.dtype("float32"): "single",
    np.dtype("bool"): "logical",
    np.dtype("int8"): "int8",
    np.dtype("uint8"): "uint8",
    np.dtype("int16"): "int16",
    np.dtype("uint16"): "uint16",
    np.dtype("int32"): "int32",
    np.dtype("uint32"): "uint32",
    np.dtype("int64"): "int64",
    np.dtype("uint64"): "uint64",
    np.dtype(np.str_): "char",
}


def mat_to_containermap(props, **_kwargs):
    """Converts MATLAB container.Map to Python dictionary"""
    comps = props.get("serialization", None)
    if comps is None:
        return props

    ks = comps[0, 0]["keys"]
    vals = comps[0, 0]["values"]

    result = {}
    for i in range(ks.shape[1]):
        key = ks[0, i].item()
        val = vals[0, i]
        result[key] = val

    return MatlabContainerMap(result)


# # def mat_to_dictionary(props, **_kwargs):
# #     """Converts MATLAB dictionary to Python list of tuples"""
# #     # List of tuples as Key-Value pairs can be any datatypes

# #     comps = props.get("data", None)
# #     if comps is None:
# #         return props

# #     ver = int(comps[0, 0]["Version"].item())
# #     if ver != MAT_DICT_VERSION:
# #         warnings.warn(
# #             f"mat_to_dictionary: Only v{MAT_DICT_VERSION} MATLAB dictionaries are supported. Got v{ver}",
# #             UserWarning,
# #         )
# #         return props

# #     ks = comps[0, 0]["Key"]
# #     vals = comps[0, 0]["Value"]

# #     return (ks, vals)


# def dictionary_to_mat(props):
#     """Converts a Python dictionary to MATLAB dictionary"""
#     if not (isinstance(props, tuple) and len(props) == 2):
#         raise TypeError("Expected tuple of (key, value)")

#     keys, values = props

#     def is_valid_array(obj):
#         return (
#             isinstance(obj, np.ndarray) or getattr(obj, "classname", None) == "string"
#         )

#     if not is_valid_array(keys) or not is_valid_array(values):
#         raise TypeError(
#             "Keys must be a numpy array or MatioOpaque with classname='string'"
#         )

#     dtype = [
#         ("Version", object),
#         ("IsKeyCombined", object),
#         ("IsValueCombined", object),
#         ("Key", object),
#         ("Value", object),
#     ]
#     data_arr = np.empty((1, 1), dtype=dtype)

#     data_arr["Key"][0, 0] = keys
#     data_arr["Value"][0, 0] = values
#     data_arr["Version"][0, 0] = np.uint64(MAT_DICT_VERSION)
#     data_arr["IsKeyCombined"][0, 0] = np.bool_(True)
#     data_arr["IsValueCombined"][0, 0] = np.bool_(True)

#     prop_map = {
#         "data": data_arr,
#     }

#     return prop_map


def detect_dtype_uniformity(vals):
    """Set uniformity and valueType for container.Map"""

    value_dtypes = set()
    val_type = None
    for v in vals:
        if not isinstance(v, np.ndarray):
            uniformity = np.bool_(False)
            val_type = "any"
            break
        if v.dtype.kind in ("U", "S"):
            value_dtypes.add(np.dtype(np.str_))
        else:
            value_dtypes.add(v.dtype)

    if val_type is None:
        if len(value_dtypes) == 1:
            uniformity = True
            val_type = map_dtype_map.get(next(iter(value_dtypes)), "any")
        else:
            uniformity = np.bool_(False)
            val_type = "any"

    return uniformity, val_type


def containermap_to_mat(props):
    """Converts a Python dictionary to MATLAB container.Map"""

    keys = list(props.keys())
    vals = list(props.values())

    val_arr = np.empty((1, len(vals)), dtype=object)
    keys_arr = np.empty((1, len(keys)), dtype=object)
    val_arr[0, :] = vals
    keys_arr[0, :] = keys

    if all(isinstance(k, str) for k in keys):
        key_type = "char"
    elif all(isinstance(k, int) for k in keys):
        key_type = "uint64"
        # Defaulting to highest precision as Python doesn't differentiate
    else:
        key_type = "double"

    uniformity, value_type = detect_dtype_uniformity(vals)

    ser_dtype = [
        ("keys", object),
        ("values", object),
        ("uniformity", object),
        ("keyType", object),
        ("valueType", object),
    ]
    serialization = np.empty((1, 1), dtype=ser_dtype)
    serialization[0, 0] = (keys_arr, val_arr, uniformity, key_type, value_type)
    prop_map = {
        "serialization": serialization,
    }
    return prop_map



import os
import tempfile

import numpy as np
import pytest

from matio import load_from_mat, save_to_mat
from matio.utils.matclass import MatlabFunction, MatlabOpaque

files = [
    ("test_function_handles_v7.mat", "v7"),
    ("test_function_handles_v73.mat", "v7.3"),
]


@pytest.mark.parametrize("filename, version", files)
class TestLoadMatlabFunctionHandles:

    def test_fh_simple(self, filename, version):
        """Test reading function handle simple from MAT-file"""
        file_path = os.path.join(os.path.dirname(__file__), "data", filename)
        mdict = load_from_mat(file_path, variable_names=["builtin_fh"])
        assert "builtin_fh" in mdict
        assert isinstance(mdict["builtin_fh"], MatlabFunction)

        fh_type = np.array(["simple"])
        np.testing.assert_array_equal(
            mdict["builtin_fh"][0, 0]["function_handle"][0, 0]["type"],
            fh_type,
            strict=True,
        )

    def test_fh_simple_custom(self, filename, version):
        """Test reading function handle simple custom from MAT-file"""
        file_path = os.path.join(os.path.dirname(__file__), "data", filename)
        mdict = load_from_mat(file_path, variable_names=["custom_fh"])
        assert "custom_fh" in mdict
        assert isinstance(mdict["custom_fh"], MatlabFunction)

        fh_type = np.array(["simple"])
        np.testing.assert_array_equal(
            mdict["custom_fh"][0, 0]["function_handle"][0, 0]["type"],
            fh_type,
            strict=True,
        )

    def test_fh_nested(self, filename, version):
        """Test reading function handle nested from MAT-file"""
        file_path = os.path.join(os.path.dirname(__file__), "data", filename)
        mdict = load_from_mat(file_path, variable_names=["nested_fh"])
        assert "nested_fh" in mdict
        assert isinstance(mdict["nested_fh"], MatlabFunction)

        fh_type = np.array(["nested"])
        np.testing.assert_array_equal(
            mdict["nested_fh"][0, 0]["function_handle"][0, 0]["type"],
            fh_type,
            strict=True,
        )
        assert isinstance(
            mdict["nested_fh"][0, 0]["function_handle"][0, 0]["workspace"], MatlabOpaque
        )

    def test_fh_classmethod(self, filename, version):
        """Test reading function handle class method from MAT-file"""
        file_path = os.path.join(os.path.dirname(__file__), "data", filename)
        mdict = load_from_mat(file_path, variable_names=["class_fh"])
        assert "class_fh" in mdict
        assert isinstance(mdict["class_fh"], MatlabFunction)

        fh_type = np.array(["anonymous"])
        np.testing.assert_array_equal(
            mdict["class_fh"][0, 0]["function_handle"][0, 0]["type"],
            fh_type,
            strict=True,
        )
        assert isinstance(
            mdict["class_fh"][0, 0]["function_handle"][0, 0]["workspace"], MatlabOpaque
        )

    def test_fh_anonymous(self, filename, version):
        """Test reading function handle anonymous from MAT-file"""
        file_path = os.path.join(os.path.dirname(__file__), "data", filename)
        mdict = load_from_mat(file_path, variable_names=["anonymous_fh"])
        assert "anonymous_fh" in mdict
        assert isinstance(mdict["anonymous_fh"], MatlabFunction)

        fh_type = np.array(["anonymous"])
        np.testing.assert_array_equal(
            mdict["anonymous_fh"][0, 0]["function_handle"][0, 0]["type"],
            fh_type,
            strict=True,
        )
        assert isinstance(
            mdict["anonymous_fh"][0, 0]["function_handle"][0, 0]["workspace"],
            MatlabOpaque,
        )


@pytest.mark.parametrize("filename, version", files)
class TestSaveMatlabFunctionHandles:

    def test_fh_simple(self, filename, version):
        """Test writing function handle simple to MAT-file"""
        file_path = os.path.join(os.path.dirname(__file__), "data", filename)
        mdict = load_from_mat(file_path, variable_names=["builtin_fh"])

        with tempfile.NamedTemporaryFile(suffix=".mat", delete=False) as tmpfile:
            temp_file_path = tmpfile.name

        try:
            save_to_mat(temp_file_path, mdict, version=version)
            mload = load_from_mat(temp_file_path, variable_names=["builtin_fh"])

            assert set(mload["builtin_fh"].dtype.names) == set(
                mdict["builtin_fh"].dtype.names
            )

        finally:
            if os.path.exists(temp_file_path):
                os.remove(temp_file_path)

    def test_fh_nested(self, filename, version):
        """Test writing function handle nested to MAT-file"""
        file_path = os.path.join(os.path.dirname(__file__), "data", filename)
        mdict = load_from_mat(file_path, variable_names=["nested_fh"])

        with tempfile.NamedTemporaryFile(suffix=".mat", delete=False) as tmpfile:
            temp_file_path = tmpfile.name

        try:
            save_to_mat(temp_file_path, mdict, version=version)
            mload = load_from_mat(temp_file_path, variable_names=["nested_fh"])

            assert set(mload["nested_fh"].dtype.names) == set(
                mdict["nested_fh"].dtype.names
            )

        finally:
            if os.path.exists(temp_file_path):
                os.remove(temp_file_path)

    def test_fh_classmethod(self, filename, version):
        """Test writing function handle class method to MAT-file"""
        file_path = os.path.join(os.path.dirname(__file__), "data", filename)
        mdict = load_from_mat(file_path, variable_names=["class_fh"])

        with tempfile.NamedTemporaryFile(suffix=".mat", delete=False) as tmpfile:
            temp_file_path = tmpfile.name

        try:
            save_to_mat(temp_file_path, mdict, version=version)
            mload = load_from_mat(temp_file_path, variable_names=["class_fh"])

            assert set(mload["class_fh"].dtype.names) == set(
                mdict["class_fh"].dtype.names
            )

        finally:
            if os.path.exists(temp_file_path):
                os.remove(temp_file_path)

    def test_fh_anonymous(self, filename, version):
        """Test writing function handle anonymous to MAT-file"""
        file_path = os.path.join(os.path.dirname(__file__), "data", filename)
        mdict = load_from_mat(file_path, variable_names=["anonymous_fh"])

        with tempfile.NamedTemporaryFile(suffix=".mat", delete=False) as tmpfile:
            temp_file_path = tmpfile.name

        try:
            save_to_mat(temp_file_path, mdict, version=version)
            mload = load_from_mat(temp_file_path, variable_names=["anonymous_fh"])

            assert set(mload["anonymous_fh"].dtype.names) == set(
                mdict["anonymous_fh"].dtype.names
            )

        finally:
            if os.path.exists(temp_file_path):
                os.remove(temp_file_path)

    def test_fh_simple_custom(self, filename, version):
        """Test writing function handle simple custom to MAT-file"""
        file_path = os.path.join(os.path.dirname(__file__), "data", filename)
        mdict = load_from_mat(file_path, variable_names=["custom_fh"])

        with tempfile.NamedTemporaryFile(suffix=".mat", delete=False) as tmpfile:
            temp_file_path = tmpfile.name

        try:
            save_to_mat(temp_file_path, mdict, version=version)
            mload = load_from_mat(temp_file_path, variable_names=["custom_fh"])

            assert set(mload["custom_fh"].dtype.names) == set(
                mdict["custom_fh"].dtype.names
            )

        finally:
            if os.path.exists(temp_file_path):
                os.remove(temp_file_path)



import os
import tempfile

import numpy as np
import pytest

from matio import load_from_mat, save_to_mat

files = [("test_string_v7.mat", "v7"), ("test_string_v73.mat", "v7.3")]


@pytest.mark.parametrize("filename, version", files)
class TestLoadMatlabString:

    def test_string_scalar(self, filename, version):
        """Test reading string scalar data from MAT-file"""
        file_path = os.path.join(os.path.dirname(__file__), "data", filename)
        mdict = load_from_mat(file_path, variable_names=["string_scalar"])
        assert "string_scalar" in mdict

        str_scalar = np.array(
            [["Hello"]], dtype=np.dtypes.StringDType(na_object=np.nan)
        )
        np.testing.assert_array_equal(mdict["string_scalar"], str_scalar, strict=True)

    def test_string_array(self, filename, version):
        """Test reading string array data from MAT-file"""
        file_path = os.path.join(os.path.dirname(__file__), "data", filename)
        mdict = load_from_mat(file_path, variable_names=["string_array"])
        assert "string_array" in mdict

        str_array = np.array(
            ["Apple", "Banana", "Cherry", "Date", "Fig", "Grapes"],
            dtype=np.dtypes.StringDType(na_object=np.nan),
        ).reshape(2, 3)
        np.testing.assert_array_equal(mdict["string_array"], str_array, strict=True)

    def test_string_empty(self, filename, version):
        """Test reading empty string data from MAT-file"""
        file_path = os.path.join(os.path.dirname(__file__), "data", filename)
        mdict = load_from_mat(file_path, variable_names=["string_empty"])
        assert "string_empty" in mdict

        str_empty = np.array([[""]], dtype=np.dtypes.StringDType(na_object=np.nan))
        np.testing.assert_array_equal(mdict["string_empty"], str_empty, strict=True)


@pytest.mark.parametrize("filename, version", files)
class TestSaveMatlabString:

    def test_string_scalar(self, filename, version):
        """Test writing string scalar to MAT-file"""
        file_path = os.path.join(os.path.dirname(__file__), "data", filename)
        mdict = load_from_mat(file_path, variable_names=["string_scalar"])

        with tempfile.NamedTemporaryFile(suffix=".mat", delete=False) as tmpfile:
            temp_file_path = tmpfile.name

        try:
            save_to_mat(temp_file_path, mdict, version=version)
            mload = load_from_mat(temp_file_path, variable_names=["string_scalar"])

            np.testing.assert_array_equal(
                mdict["string_scalar"], mload["string_scalar"], strict=True
            )

        finally:
            if os.path.exists(temp_file_path):
                os.remove(temp_file_path)

    def test_string_array(self, filename, version):
        """Test writing string array to MAT-file"""
        file_path = os.path.join(os.path.dirname(__file__), "data", filename)
        mdict = load_from_mat(file_path, variable_names=["string_array"])

        with tempfile.NamedTemporaryFile(suffix=".mat", delete=False) as tmpfile:
            temp_file_path = tmpfile.name

        try:
            save_to_mat(temp_file_path, mdict, version=version)
            mload = load_from_mat(temp_file_path, variable_names=["string_array"])

            np.testing.assert_array_equal(
                mdict["string_array"], mload["string_array"], strict=True
            )

        finally:
            if os.path.exists(temp_file_path):
                os.remove(temp_file_path)

    def test_string_empty(self, filename, version):
        """Test writing empty string to MAT-file"""
        file_path = os.path.join(os.path.dirname(__file__), "data", filename)
        mdict = load_from_mat(file_path, variable_names=["string_empty"])

        with tempfile.NamedTemporaryFile(suffix=".mat", delete=False) as tmpfile:
            temp_file_path = tmpfile.name

        try:
            save_to_mat(temp_file_path, mdict, version=version)
            mload = load_from_mat(temp_file_path, variable_names=["string_empty"])

            np.testing.assert_array_equal(
                mdict["string_empty"], mload["string_empty"], strict=True
            )

        finally:
            if os.path.exists(temp_file_path):
                os.remove(temp_file_path)



import os
import tempfile

import numpy as np
import pytest

from matio import load_from_mat, save_to_mat
from matio.utils.matclass import MatConvertWarning

files = [("test_time_v7.mat", "v7"), ("test_time_v73.mat", "v7.3")]

# Notes
# During load, fallback subdtype is 'ms'
# However, during save, fallback is 's' as required by MATLAB


@pytest.mark.parametrize("filename, version", files)
class TestLoadMatlabDuration:

    def test_duration_seconds(self, filename, version):
        """Test reading duration seconds from MAT-file"""
        file_path = os.path.join(os.path.dirname(__file__), "data", filename)
        mdict = load_from_mat(file_path, variable_names=["dur_s"])
        assert "dur_s" in mdict

        dur_seconds = np.array([5], dtype="timedelta64[s]").reshape(1, 1)

        np.testing.assert_array_equal(mdict["dur_s"], dur_seconds, strict=True)

    def test_duration_minutes(self, filename, version):
        """Test reading duration minutes data from MAT-file"""
        file_path = os.path.join(os.path.dirname(__file__), "data", filename)
        mdict = load_from_mat(file_path, variable_names=["dur_m"])
        assert "dur_m" in mdict

        dur_minutes = np.array([5], dtype="timedelta64[m]").reshape(1, 1)

        np.testing.assert_array_equal(mdict["dur_m"], dur_minutes, strict=True)

    def test_duration_hours(self, filename, version):
        """Test reading duration hours from MAT-file"""
        file_path = os.path.join(os.path.dirname(__file__), "data", filename)
        mdict = load_from_mat(file_path, variable_names=["dur_h"])
        assert "dur_h" in mdict

        dur_hours = np.array([5], dtype="timedelta64[h]").reshape(1, 1)

        np.testing.assert_array_equal(mdict["dur_h"], dur_hours, strict=True)

    def test_duration_days(self, filename, version):
        """Test reading duration days from MAT-file"""
        file_path = os.path.join(os.path.dirname(__file__), "data", filename)
        mdict = load_from_mat(file_path, variable_names=["dur_days"])
        assert "dur_days" in mdict

        dur_days = np.array([5], dtype="timedelta64[D]").reshape(1, 1)

        np.testing.assert_array_equal(mdict["dur_days"], dur_days, strict=True)

    def test_duration_years(self, filename, version):
        """Test reading duration years from MAT-file"""
        file_path = os.path.join(os.path.dirname(__file__), "data", filename)
        mdict = load_from_mat(file_path, variable_names=["dur_years"])
        assert "dur_years" in mdict

        dur_years = np.array([1, 2, 3], dtype="timedelta64[Y]").reshape(1, 3)

        np.testing.assert_array_equal(mdict["dur_years"], dur_years, strict=True)

    def test_duration_hms(self, filename, version):
        """Test reading duration hh:mm:ss from MAT-file"""
        file_path = os.path.join(os.path.dirname(__file__), "data", filename)
        with pytest.warns(MatConvertWarning, match="mat_to_duration: Unknown format"):
            mdict = load_from_mat(file_path, variable_names=["dur_hms"])
            assert "dur_hms" in mdict

            dur_vec = (
                (
                    np.timedelta64(1, "h")
                    + np.timedelta64(2, "m")
                    + np.timedelta64(3, "s")
                )
                .astype("timedelta64[ms]")
                .reshape(1, 1)
            )
            np.testing.assert_array_equal(mdict["dur_hms"], dur_vec, strict=True)

    def test_duration_array(self, filename, version):
        """Test reading duration array from MAT-file"""
        file_path = os.path.join(os.path.dirname(__file__), "data", filename)
        mdict = load_from_mat(file_path, variable_names=["dur_array"])
        assert "dur_array" in mdict

        dur_array = np.array([10, 20, 30, 40, 50, 60], dtype="timedelta64[s]").reshape(
            2, 3
        )
        np.testing.assert_array_equal(mdict["dur_array"], dur_array, strict=True)

    def test_duration_empty(self, filename, version):
        """Test reading duration empty from MAT-file"""
        file_path = os.path.join(os.path.dirname(__file__), "data", filename)

        with pytest.warns(MatConvertWarning, match="mat_to_duration: Unknown format"):
            mdict = load_from_mat(file_path, variable_names=["dur_empty"])

        assert "dur_empty" in mdict

        dur_empty = np.empty((0, 0), dtype="timedelta64[ms]")
        np.testing.assert_array_equal(mdict["dur_empty"], dur_empty, strict=True)


@pytest.mark.parametrize("filename, version", files)
class TestSaveMatlabDuration:

    def test_duration_seconds(self, filename, version):
        """Test writing duration seconds to MAT-file"""
        file_path = os.path.join(os.path.dirname(__file__), "data", filename)
        mdict = load_from_mat(file_path, variable_names=["dur_s"])

        with tempfile.NamedTemporaryFile(suffix=".mat", delete=False) as tmpfile:
            temp_file_path = tmpfile.name

        try:
            save_to_mat(temp_file_path, mdict, version=version)
            mload = load_from_mat(temp_file_path, variable_names=["dur_s"])

            np.testing.assert_array_equal(mdict["dur_s"], mload["dur_s"], strict=True)

        finally:
            if os.path.exists(temp_file_path):
                os.remove(temp_file_path)

    def test_duration_minutes(self, filename, version):
        """Test writing duration minutes to MAT-file"""
        file_path = os.path.join(os.path.dirname(__file__), "data", filename)
        mdict = load_from_mat(file_path, variable_names=["dur_m"])

        with tempfile.NamedTemporaryFile(suffix=".mat", delete=False) as tmpfile:
            temp_file_path = tmpfile.name

        try:
            save_to_mat(temp_file_path, mdict, version=version)

            mload = load_from_mat(temp_file_path, variable_names=["dur_m"])

            np.testing.assert_array_equal(mdict["dur_m"], mload["dur_m"], strict=True)

        finally:
            if os.path.exists(temp_file_path):
                os.remove(temp_file_path)

    def test_duration_hours(self, filename, version):
        """Test writing duration hours to MAT-file"""
        file_path = os.path.join(os.path.dirname(__file__), "data", filename)
        mdict = load_from_mat(file_path, variable_names=["dur_h"])

        with tempfile.NamedTemporaryFile(suffix=".mat", delete=False) as tmpfile:
            temp_file_path = tmpfile.name

        try:
            save_to_mat(temp_file_path, mdict, version=version)
            mload = load_from_mat(temp_file_path, variable_names=["dur_h"])

            np.testing.assert_array_equal(mdict["dur_h"], mload["dur_h"], strict=True)

        finally:
            if os.path.exists(temp_file_path):
                os.remove(temp_file_path)

    def test_duration_days(self, filename, version):
        """Test writing duration days to MAT-file"""
        file_path = os.path.join(os.path.dirname(__file__), "data", filename)
        mdict = load_from_mat(file_path, variable_names=["dur_days"])

        with tempfile.NamedTemporaryFile(suffix=".mat", delete=False) as tmpfile:
            temp_file_path = tmpfile.name

        try:
            save_to_mat(temp_file_path, mdict, version=version)
            mload = load_from_mat(temp_file_path, variable_names=["dur_days"])

            np.testing.assert_array_equal(
                mdict["dur_days"], mload["dur_days"], strict=True
            )

        finally:
            if os.path.exists(temp_file_path):
                os.remove(temp_file_path)

    def test_duration_years(self, filename, version):
        """Test writing duration years to MAT-file"""
        file_path = os.path.join(os.path.dirname(__file__), "data", filename)
        mdict = load_from_mat(file_path, variable_names=["dur_years"])

        with tempfile.NamedTemporaryFile(suffix=".mat", delete=False) as tmpfile:
            temp_file_path = tmpfile.name

        try:
            save_to_mat(temp_file_path, mdict, version=version)
            mload = load_from_mat(temp_file_path, variable_names=["dur_years"])

            np.testing.assert_array_equal(
                mdict["dur_years"], mload["dur_years"], strict=True
            )

        finally:
            if os.path.exists(temp_file_path):
                os.remove(temp_file_path)

    def test_duration_array(self, filename, version):
        """Test writing duration array to MAT-file"""
        file_path = os.path.join(os.path.dirname(__file__), "data", filename)
        mdict = load_from_mat(file_path, variable_names=["dur_array"])

        with tempfile.NamedTemporaryFile(suffix=".mat", delete=False) as tmpfile:
            temp_file_path = tmpfile.name

        try:
            save_to_mat(temp_file_path, mdict, version=version)
            mload = load_from_mat(temp_file_path, variable_names=["dur_array"])

            np.testing.assert_array_equal(
                mdict["dur_array"], mload["dur_array"], strict=True
            )

        finally:
            if os.path.exists(temp_file_path):
                os.remove(temp_file_path)

    def test_duration_empty(self, filename, version):
        """Test writing empty duration to MAT-file"""
        file_path = os.path.join(os.path.dirname(__file__), "data", filename)

        with pytest.warns(MatConvertWarning, match="mat_to_duration: Unknown format"):
            mdict = load_from_mat(file_path, variable_names=["dur_empty"])

        with tempfile.NamedTemporaryFile(suffix=".mat", delete=False) as tmpfile:
            temp_file_path = tmpfile.name

        try:
            with pytest.warns(MatConvertWarning, match="duration_to_mat"):
                save_to_mat(temp_file_path, mdict, version=version)

            mload = load_from_mat(temp_file_path, variable_names=["dur_empty"])

            np.testing.assert_array_equal(
                mdict["dur_empty"].astype("timedelta64[s]"),
                mload["dur_empty"],
                strict=True,
            )

        finally:
            if os.path.exists(temp_file_path):
                os.remove(temp_file_path)

    def test_duration_hms(self, filename, version):
        """Test writing duration h:m:s to MAT-file"""
        file_path = os.path.join(os.path.dirname(__file__), "data", filename)
        with pytest.warns(MatConvertWarning, match="mat_to_duration: Unknown format"):
            mdict = load_from_mat(file_path, variable_names=["dur_hms"])

        with tempfile.NamedTemporaryFile(suffix=".mat", delete=False) as tmpfile:
            temp_file_path = tmpfile.name

        try:

            with pytest.warns(MatConvertWarning, match="duration_to_mat"):
                save_to_mat(temp_file_path, mdict, version=version)

            mload = load_from_mat(temp_file_path, variable_names=["dur_hms"])

            np.testing.assert_array_equal(
                mdict["dur_hms"].astype("timedelta64[s]"), mload["dur_hms"], strict=True
            )

        finally:
            if os.path.exists(temp_file_path):
                os.remove(temp_file_path)



"""MAT-file version 7.3 (HDF5) writer."""

import string
import sys
import warnings
from itertools import count, product

import h5py
import numpy as np
from scipy.sparse import issparse

from matio.subsystem import MatSubsystem
from matio.utils.matclass import (
    MCOS_SUBSYSTEM_CLASS,
    EmptyMatStruct,
    IntegerDecodingHint,
    MatlabCanonicalEmpty,
    MatlabClasses,
    MatlabEnumerationArray,
    MatlabFunction,
    MatlabObject,
    MatlabOpaque,
    MatlabOpaqueArray,
    MatWriteWarning,
    ObjectDecodingHint,
)
from matio.utils.matheaders import (
    MAT_HDF_ATTRS,
    MAT_HDF_COMPRESSION,
    MAT_HDF_COMPRESSION_OPTS,
    MAT_HDF_REFS_GROUP,
    MAT_HDF_SUBSYS_GROUP,
    MAT_HDF_USER_BLOCK_BYTES,
    MAT_HDF_VERSION,
    write_file_header,
)
from matio.utils.matutils import mat_numeric, strings_to_chars, to_writeable

SYS_BYTE_ORDER = "<" if sys.byteorder == "little" else ">"


def savemat7(file_path, mdict, global_vars, saveobj_classes, oned_as):
    """Write data to MAT-5file."""

    with h5py.File(file_path, "w", userblock_size=MAT_HDF_USER_BLOCK_BYTES) as f:
        MW7 = MatWrite7(f, oned_as=oned_as)
        MW7.subsystem = MatSubsystem(
            byte_order=SYS_BYTE_ORDER, oned_as=oned_as, saveobj_classes=saveobj_classes
        )
        MW7.subsystem.init_save()
        MW7.put_variables(mdict, global_vars)

        subsystem = MW7.subsystem.set_subsystem()
        if subsystem is not None:
            MW7.write_struct_array(f, MAT_HDF_SUBSYS_GROUP, subsystem)

    with open(file_path, "r+b") as f:
        f.seek(0)
        write_file_header(f, version=MAT_HDF_VERSION)

    return


class MatWrite7:
    """Writes MAT-file version 7.3 (HDF5) files."""

    def __init__(self, h5file, oned_as="col"):
        self.h5file = h5file
        self.oned_as = oned_as
        self.subsystem = None
        self.refs_group = MAT_HDF_REFS_GROUP
        self._name_gen = self._matlab_refname_generator()

    def _matlab_refname_generator(self):
        """Generates MATLAB style refs subgroup names."""
        yield from string.ascii_lowercase
        yield from string.ascii_uppercase
        alphabet = string.ascii_lowercase + string.ascii_uppercase
        for size in count(2):
            for letters in product(alphabet, repeat=size):
                yield "".join(letters)

    def _next_refname(self):
        """Gets the next MATLAB style refs subgroup name."""
        return next(self._name_gen)

    def get_empty_array(self, shape):
        """MATLAB stores empty arrays as uint64 containing shape"""
        return np.array(shape, dtype=np.uint64)

    def add_empty_attribute(self, dset):
        """Adds MATLAB empty attribute to the dataset."""
        dset.attrs.create(MAT_HDF_ATTRS.EMPTY, np.int32(1))

    def add_classname_attr(self, dset, classname):
        """Adds MATLAB class name attribute to the dataset."""
        dset.attrs.create(MAT_HDF_ATTRS.CLASS, np.bytes_(classname))

    def add_int_decode_attr(self, dset, int_decode):
        """Adds MATLAB integer decoding hint attribute to the dataset."""
        dset.attrs.create(MAT_HDF_ATTRS.INT_DECODE, np.int32(int_decode))

    def add_object_decode_attr(self, dset, object_decode):
        """Adds MATLAB object decoding hint attribute to the dataset."""
        dset.attrs.create(MAT_HDF_ATTRS.OBJECT_DECODE, np.int32(object_decode))

    def add_sparse_attr(self, dset, nrows):
        """Adds MATLAB sparse nnz attribute to the dataset."""
        dset.attrs.create(MAT_HDF_ATTRS.SPARSE, np.int32(nrows))

    def write_numeric_dset(self, parent, var_name, data):
        """Writes a numeric dataset to the HDF5 file."""

        data, classname, int_decode = mat_numeric(data, version=MAT_HDF_VERSION)
        if data.size == 0:
            data_empty = self.get_empty_array(data.shape)
            dset = parent.create_dataset(var_name, data=data_empty)
            self.add_empty_attribute(dset)
        else:
            dset = parent.create_dataset(
                var_name,
                data=data.T,
                compression=MAT_HDF_COMPRESSION,
                compression_opts=MAT_HDF_COMPRESSION_OPTS,
                chunks=True,
            )

        self.add_classname_attr(dset, classname)
        if int_decode is not None and int_decode == IntegerDecodingHint.LOGICAL_HINT:
            self.add_int_decode_attr(dset, int_decode)

        return dset

    def write_char_dset(self, parent, var_name, data):
        """Writes a char dataset to the HDF5 file."""

        if data.size == 0 or np.all(data == ""):
            # Empty String Array
            # MATLAB only recognizes 0-D char array
            shape = (0,) * np.max([data.ndim, 2])
            data_empty = self.get_empty_array(shape)
            dset = parent.create_dataset(var_name, data=data_empty)
            self.add_empty_attribute(dset)
        else:
            data = strings_to_chars(data)
            data = data.view(np.uint32).astype(np.uint16)
            dset = parent.create_dataset(
                var_name,
                data=data.T,
                compression=MAT_HDF_COMPRESSION,
                compression_opts=MAT_HDF_COMPRESSION_OPTS,
                chunks=True,
            )

        self.add_classname_attr(dset, MatlabClasses.CHAR)
        self.add_int_decode_attr(dset, IntegerDecodingHint.UTF16_HINT)

        return dset

    def write_cell_array(self, parent, var_name, data):
        """Writes a cell array to the HDF5 file."""

        if data.size == 0:
            data = self.get_empty_array(data.shape)
            dset = parent.create_dataset(var_name, data=data.T)
            self.add_classname_attr(dset, MatlabClasses.CELL)
            self.add_empty_attribute(dset)

        else:
            ref_array = np.empty(data.shape, dtype=h5py.ref_dtype)

            for idx, item in np.ndenumerate(data):
                dset_name = self._next_refname()
                dset = self.write_variable(dset_name, item, group=self.refs_group)
                ref_array[idx] = dset.ref

            dset = parent.create_dataset(
                var_name, data=ref_array.T, dtype=h5py.ref_dtype
            )

        self.add_classname_attr(dset, MatlabClasses.CELL)
        return dset

    def write_empty_struct(self, parent, var_name, data):
        """Writes an empty struct to the HDF5 file."""

        data = self.get_empty_array(data.shape)
        dset = parent.create_dataset(var_name, data=data)
        self.add_classname_attr(dset, MatlabClasses.STRUCT)
        self.add_empty_attribute(dset)
        return dset

    def write_struct_array(self, parent, var_name, data, object_decode=0):
        """Writes a struct array to the HDF5 file."""

        fields = data.dtype.names
        if data.size <= 1:
            struct_group = parent.create_group(var_name)
            if fields is not None:
                for field in fields:
                    self.write_variable(
                        field, data[field][0, 0], group=struct_group.name
                    )
        else:
            # Struct arrays are stored as dset of references to each struct
            struct_group = parent.create_group(var_name)
            for field in fields:
                field_refs = np.empty(data.shape, dtype=h5py.ref_dtype)
                for idx, item in np.ndenumerate(data):
                    dset_name = self._next_refname()
                    dset = self.write_variable(
                        dset_name, item[field], group=self.refs_group
                    )
                    field_refs[idx] = dset.ref
                dset = struct_group.create_dataset(
                    field, data=field_refs.T, dtype=h5py.ref_dtype
                )

        if object_decode == ObjectDecodingHint.FUNCTION_HINT:
            self.add_object_decode_attr(struct_group, object_decode)
            self.add_classname_attr(struct_group, MatlabClasses.FUNCTION)
        elif object_decode == ObjectDecodingHint.OBJECT_HINT:
            self.add_object_decode_attr(struct_group, object_decode)
            self.add_classname_attr(struct_group, data.classname)
        else:
            self.add_classname_attr(struct_group, MatlabClasses.STRUCT)

        # Add field names attribute
        # MATLAB uses an assertion to ensure sub-group/dataset names matches with
        # field names in attributes
        dt = h5py.special_dtype(vlen=np.dtype("S1"))
        matlab_fields = np.empty(shape=(len(fields),), dtype=dt)
        for i, field_name in enumerate(fields):
            matlab_fields[i] = np.array(
                [c.encode("ascii") for c in field_name], dtype="S1"
            )

        if matlab_fields.size > 0:
            struct_group.attrs.create(MAT_HDF_ATTRS.FIELDS, matlab_fields)

        return struct_group

    def write_function_handle(self, parent, var_name, data):
        """Writes a function handle to the HDF5 file."""

        if data.size == 0:
            warnings.warn(
                "Empty function handle not supported. Skipping.", MatWriteWarning
            )
            return

        dset = self.write_struct_array(
            parent, var_name, data, object_decode=ObjectDecodingHint.FUNCTION_HINT
        )
        return dset

    def write_matlab_object(self, parent, var_name, data):
        """Writes a MATLAB object to the HDF5 file."""

        if data.size == 0:
            warnings.warn(
                "Empty MATLAB object not supported. Skipping.", MatWriteWarning
            )
            return

        dset = self.write_struct_array(
            parent, var_name, data, object_decode=ObjectDecodingHint.OBJECT_HINT
        )
        return dset

    def write_sparse_array(self, parent, var_name, data):
        """Writes a sparse array to the HDF5 file."""

        A = data.tocsc()
        A.sort_indices()
        ir = A.indices.astype("uint64")
        jc = A.indptr.astype("uint64")

        data, classname, int_decode = mat_numeric(A.data, version=MAT_HDF_VERSION)

        sparse_group = parent.create_group(var_name)
        sparse_group.create_dataset(
            "jc",
            data=jc,
            compression=MAT_HDF_COMPRESSION,
            compression_opts=MAT_HDF_COMPRESSION_OPTS,
            chunks=True,
        )
        if data.size > 0:
            sparse_group.create_dataset(
                "data",
                data=data,
                compression=MAT_HDF_COMPRESSION,
                compression_opts=MAT_HDF_COMPRESSION_OPTS,
                chunks=True,
            )
            sparse_group.create_dataset(
                "ir",
                data=ir,
                compression=MAT_HDF_COMPRESSION,
                compression_opts=MAT_HDF_COMPRESSION_OPTS,
                chunks=True,
            )

        self.add_classname_attr(sparse_group, classname)
        self.add_sparse_attr(sparse_group, A.shape[0])
        if int_decode is not None and int_decode == IntegerDecodingHint.LOGICAL_HINT:
            self.add_int_decode_attr(sparse_group, int_decode)
        return sparse_group

    def write_opaque_object(self, parent, var_name, data):
        """Writes an opaque object to the HDF5 file."""

        if data.classname == MCOS_SUBSYSTEM_CLASS:
            dset = self.write_cell_array(parent, var_name, data.properties)
        else:
            if isinstance(data, MatlabEnumerationArray):
                metadata = self.subsystem.set_enumeration_metadata(data)
                dset = self.write_struct_array(
                    parent,
                    var_name,
                    metadata,
                    object_decode=ObjectDecodingHint.OPAQUE_HINT,
                )
            else:
                metadata = self.subsystem.set_object_metadata(data)
                dset = parent.create_dataset(var_name, data=metadata.T)

        classname = data.classname

        dset.attrs.create(MAT_HDF_ATTRS.CLASS, np.bytes_(classname))
        dset.attrs.create(
            MAT_HDF_ATTRS.OBJECT_DECODE, np.int32(ObjectDecodingHint.OPAQUE_HINT)
        )
        return dset

    def write_canonical_empty(self, parent, var_name, data):
        """Writes a canonical empty array to the HDF5 file."""

        data_empty = self.get_empty_array((0, 0))
        dset = parent.create_dataset(var_name, data=data_empty)
        self.add_classname_attr(dset, MatlabClasses.EMPTY)
        self.add_empty_attribute(dset)
        return dset

    def write_variable(self, var_name, data, group=None):
        """Writes a variable to the HDF5 file."""

        if group is None:
            parent = self.h5file
        else:
            parent = self.h5file.require_group(group)

        data = to_writeable(data, self.oned_as)

        if issparse(data):
            dset = self.write_sparse_array(parent, var_name, data)
        elif isinstance(data, MatlabCanonicalEmpty):
            dset = self.write_canonical_empty(parent, var_name, data)
        elif isinstance(
            data, (MatlabOpaque, MatlabOpaqueArray, MatlabEnumerationArray)
        ):
            dset = self.write_opaque_object(parent, var_name, data)
        elif isinstance(data, MatlabFunction):
            dset = self.write_function_handle(parent, var_name, data)
        elif isinstance(data, MatlabObject):
            dset = self.write_matlab_object(parent, var_name, data)
        elif isinstance(data, EmptyMatStruct):
            dset = self.write_empty_struct(parent, var_name, data)

        # At this point it must be a numpy array
        elif data.dtype.kind in ("i", "u", "f", "c", "b"):
            dset = self.write_numeric_dset(parent, var_name, data)
        elif data.dtype.kind in ("U", "S"):
            dset = self.write_char_dset(parent, var_name, data)
        elif data.dtype.hasobject:
            if data.dtype.names:
                dset = self.write_struct_array(parent, var_name, data)
            else:
                dset = self.write_cell_array(parent, var_name, data)
        else:
            warnings.warn(
                f"Data type {data.dtype} not supported for variable {var_name}. Skipping.",
                MatWriteWarning,
            )
            return None

        if group is not None:
            dset.attrs.create("H5PATH", np.bytes_(dset.name))

        return dset

    def put_variables(self, mdict, global_vars):
        """Writes variables to the HDF5 file."""

        for var_name, data in mdict.items():
            if var_name[0] == "_":
                msg = (
                    f"Starting field name with a " f"underscore ({var_name}) is ignored"
                )
                warnings.warn(msg, MatWriteWarning, stacklevel=2)
                continue

            dset = self.write_variable(var_name, data)
            if dset is not None and var_name in global_vars:
                dset.attrs.create(MAT_HDF_ATTRS.GLOBAL, np.int32(1))

        return



from collections import UserDict
from enum import IntEnum, StrEnum

import numpy as np

MCOS_SUBSYSTEM_CLASS = "FileWrapper__"
DYNAMIC_PROPERTY_PREFIX = "__dynamic_property__"
ENUMERATION_INSTANCE_TAG = "EnumerationInstanceTag"


class IntegerDecodingHint(IntEnum):
    """MATLAB integer arrays typedefs"""

    NO_INT_HINT = 0
    LOGICAL_HINT = 1
    UTF16_HINT = 2


class ObjectDecodingHint(IntEnum):
    """MATLAB object arrays typdefs"""

    NO_OBJ_HINT = 0
    FUNCTION_HINT = 1
    OBJECT_HINT = 2
    OPAQUE_HINT = 3


class PropertyType(IntEnum):
    """Enumeration for property types in FileWrapper__ data"""

    MATLAB_ENUMERATION = 0  # A string which names some internal enumeration probably
    PROPERTY_VALUE = 1
    INTEGER_VALUE = (
        2  # Mostly logical but I've seen integer values as well in fig files
    )


class OpaqueType(StrEnum):
    """Enumeration for different opaque object type systems"""

    MCOS = "MCOS"
    JAVA = "java"
    HANDLE = "handle"


class MatlabClasses(StrEnum):
    """Enumeration for different MATLAB class names"""

    CHAR = "char"
    STRUCT = "struct"
    CELL = "cell"
    FUNCTION = "function_handle"
    LOGICAL = "logical"
    DOUBLE = "double"
    SINGLE = "single"
    EMPTY = "canonical empty"


class MatReadError(Exception):
    """Exception indicating a read issue."""


class MatWriteError(Exception):
    """Exception indicating a write issue."""


class MatReadWarning(UserWarning):
    """Warning class for read issues."""


class MatWriteWarning(UserWarning):
    """Warning class for write issues."""


class MatConvertWarning(UserWarning):
    """Warning class for conversion issues."""


class MatConvertError(Exception):
    """Exception indicating a conversion issue."""


class MatlabObject(np.ndarray):
    """Subclass for a MATLAB object."""

    def __new__(cls, input_array, classname=None):
        """Create a new instance of MatlabObject"""
        obj = np.asarray(input_array).view(cls)
        obj.classname = classname
        return obj

    def __array_finalize__(self, obj):
        """Finalize the array, copying the classname."""
        self.classname = getattr(obj, "classname", None)


class MatlabFunction(np.ndarray):
    """Subclass for a MATLAB function."""

    def __new__(cls, input_array):
        """Create a new instance of MatlabFunction"""
        obj = np.asarray(input_array).view(cls)
        return obj


class MatlabOpaque:
    """Subclass for a MATLAB opaque object."""

    def __init__(
        self, properties, classname, type_system=OpaqueType.MCOS, class_alias=None
    ):
        """Create a new instance of MatlabOpaque"""
        self.properties = properties
        self.classname = classname
        self.type_system = type_system
        self.class_alias = class_alias

    def __repr__(self):
        """String representation of the object"""
        if isinstance(self.properties, tuple):
            shape = self.properties
        else:
            shape = (1, 1)
        return f"MatlabOpaque(classname={self.classname})"


class MatlabOpaqueArray(np.ndarray):
    """Subclass for a MATLAB Opaque array."""

    def __new__(cls, input_array, classname, type_system=OpaqueType.MCOS):
        """Create a new instance of MatlabMCOSArray"""
        obj = np.asarray(input_array).view(cls)
        obj.classname = classname
        obj.type_system = type_system
        return obj

    def __array_finalize__(self, obj):
        """Finalize the array, copying the classnames."""
        self.type_system = getattr(obj, "type_system", None)
        self.classname = getattr(obj, "classname", None)


class MatlabEnumerationArray(np.ndarray):
    """Subclass for a MATLAB Enumeration array."""

    def __new__(cls, input_array, classname, type_system=OpaqueType.MCOS):
        """Create a new instance of MatlabEnumerationArray"""
        obj = np.asarray(input_array).view(cls)
        obj.classname = classname
        obj.type_system = type_system
        return obj

    def __array_finalize__(self, obj):
        """Finalize the array, copying the classnames."""
        self.type_system = getattr(obj, "type_system", None)
        self.classname = getattr(obj, "classname", None)


class MatlabOpaqueProperty(np.ndarray):
    """Subclass for special Matlab Opaque properties."""

    def __new__(cls, input_array, ptype):
        """Create a new instance of MatlabObject"""
        obj = np.asarray(input_array).view(cls)
        obj.ptype = ptype
        return obj

    def __array_finalize__(self, obj):
        """Finalize the array, copying the classname."""
        self.ptype = getattr(obj, "ptype", None)


class MatlabContainerMap(UserDict):
    """Class to represent a MATLAB containers.Map object."""

    classname = "container.Map"


class EmptyMatStruct(np.ndarray):
    """Class to represent an empty MATLAB struct."""

    def __new__(cls, input_array):
        """Create a new instance of EmptyMatStruct"""
        obj = np.asarray(input_array).view(cls)
        return obj


class MatlabCanonicalEmpty:
    """Class to represent a canonical empty MATLAB array."""



"""MATLAB MAT-file version 7.3 (HDF5) reader."""

import warnings

import h5py
import numpy as np
from scipy.sparse import csc_array

from matio.subsystem import MatSubsystem
from matio.utils.matclass import (
    ENUMERATION_INSTANCE_TAG,
    MCOS_SUBSYSTEM_CLASS,
    EmptyMatStruct,
    IntegerDecodingHint,
    MatlabCanonicalEmpty,
    MatlabFunction,
    MatlabObject,
    MatReadWarning,
    ObjectDecodingHint,
)
from matio.utils.matconvert import guess_type_system
from matio.utils.matheaders import (
    MAT_HDF_ATTRS,
    MAT_HDF_REFS_GROUP,
    MAT_HDF_SUBSYS_GROUP,
    MCOS_MAGIC_NUMBER,
)
from matio.utils.matutils import (
    chars_to_strings,
    matlab_class_to_dtype,
    shape_from_metadata,
)


def loadmat7(
    file_path,
    byte_order,
    variable_names,
    raw_data,
    add_table_attrs,
):
    """Reads MAT-file version 7.3 (HDF5) files."""

    with h5py.File(file_path, "r") as f:
        subsystem = read_subsystem(f, byte_order, raw_data, add_table_attrs)
        mat_reader = MatRead7(f, subsystem)
        matfile_dict = mat_reader.get_variables(variable_names)

    return matfile_dict


def whosmat7(file_path):
    """Lists variables in MAT-file version 7.3 (HDF5) files."""

    with h5py.File(file_path, "r") as f:
        vars = []
        for var in f:
            if var in (MAT_HDF_REFS_GROUP, MAT_HDF_SUBSYS_GROUP):
                continue

            class_name = f[var].attrs.get(MAT_HDF_ATTRS.CLASS, None)
            object_decode = f[var].attrs.get(MAT_HDF_ATTRS.OBJECT_DECODE, -1)
            is_empty = f[var].attrs.get(MAT_HDF_ATTRS.EMPTY, 0)
            is_sparse = f[var].attrs.get(MAT_HDF_ATTRS.SPARSE, -1)

            if class_name is not None:
                class_name = class_name.decode("ascii")

            if is_empty:
                shape = tuple(int(x) for x in f[var][()])
            elif is_sparse >= 0:
                ncols = f[var]["jc"].size - 1
                shape = (int(is_sparse), ncols)
                class_name = "sparse"
            elif object_decode > 0:
                if object_decode == ObjectDecodingHint.OPAQUE_HINT:
                    if isinstance(f[var], h5py.Dataset):
                        objmetadata = f[var][()].T
                        shape = shape_from_metadata(objmetadata)
                    else:
                        if MAT_HDF_ATTRS.FIELDS in f[var].attrs:
                            # Enumeration Array
                            shape = f[var]["ValueIndices"].shape[::-1]
                        else:
                            warnings.warn(
                                f"Cannot determine shape of opaque object {var}",
                                MatReadWarning,
                            )
                            shape = ()
                else:
                    shape = (1, 1)
            elif class_name == "struct":
                keys = list(f[var].keys())
                if len(keys) == 0:
                    # Struct with no fields
                    shape = (1, 1)
                f[var][keys[0]]
                if isinstance(f[var][keys[0]], h5py.Dataset) and f[var][
                    keys[0]
                ].dtype == np.dtype("O"):
                    # Struct array
                    shape = f[var][keys[0]].shape[::-1]
                else:
                    # Scalar struct
                    shape = (1, 1)
            else:
                shape = f[var].shape[::-1]

            vars.append((var, shape, class_name))

        return vars


class MatRead7:
    """Reads MAT-file version 7.3 (HDF5) files."""

    def __init__(self, file_stream, subsystem=None):
        """Initializes the MAT-file reader.
        Parameters
        ----------
            file_stream : h5py.File object
            raw_data : bool, optional
            add_table_attrs : bool, optional
            chars_as_strings : bool, optional
        """
        self.h5stream = file_stream
        self.subsystem = subsystem

    def read_numeric(self, obj, is_empty=0):
        """Reads MATLAB numeric arrays from the v7.3 MAT-file."""

        int_decode = obj.attrs.get(MAT_HDF_ATTRS.INT_DECODE, None)

        if is_empty:
            obj_class = obj.attrs.get(MAT_HDF_ATTRS.CLASS, None)
            dtype = matlab_class_to_dtype(obj_class, obj.dtype)
            return np.empty(shape=obj[()], dtype=dtype)

        arr = obj[()]
        if arr.dtype.names:
            # complex number
            real = arr["real"]
            imag = arr["imag"]
            arr = np.empty(shape=arr.shape, dtype=np.complex128)
            arr.real = real
            arr.imag = imag

        if int_decode == IntegerDecodingHint.LOGICAL_HINT:
            arr = arr.astype(np.bool_)

        return arr.T

    def read_char(self, obj, is_empty=0):
        """Decodes MATLAB char arrays from the v7.3 MAT-file."""

        decode_type = obj.attrs.get(MAT_HDF_ATTRS.INT_DECODE, None)
        raw = obj[()].T

        if is_empty:
            return chars_to_strings(np.empty(shape=raw, dtype=np.str_))

        if decode_type == IntegerDecodingHint.UTF16_HINT:
            codec = "utf-16"
        else:
            warnings.warn(
                f"MatRead7:read_char:{MAT_HDF_ATTRS.INT_DECODE}={decode_type} not supported. "
                "This may lead to unexpected behaviour",
                MatReadWarning,
            )
            codec = "utf-8"

        decoded_arr = np.array(list(raw.tobytes().decode(codec))).reshape(raw.shape)
        return chars_to_strings(decoded_arr)

    def is_struct_matrix(self, hdf5_group):
        """Check if the HDF5 struct group is a struct matrix or scalar"""

        # Scalar structs are stored directly as members of a group (can be nested)
        # Struct arrays are stored as datasets of HDF5 references
        for key in hdf5_group:
            obj = hdf5_group[key]
            if isinstance(obj, h5py.Group):
                return False

            if isinstance(obj, h5py.Dataset):
                class_name = obj.attrs.get(MAT_HDF_ATTRS.CLASS, None)
                if class_name is not None:
                    return False

        return True

    def read_struct(self, obj, is_empty=0):
        """Reads MATLAB struct arrays from the v7.3 MAT-file."""

        if is_empty:
            return EmptyMatStruct(np.empty(shape=obj[()], dtype=object))

        fields = list(obj.keys())
        field_order = obj.attrs.get(MAT_HDF_ATTRS.FIELDS, None)
        if field_order is not None:
            # For maximum compatibility with scipy.io
            fields = ["".join(x.astype(str)) for x in field_order]

        if self.is_struct_matrix(obj):
            is_scalar = False
            shape = next(iter(obj.values())).shape
        else:
            is_scalar = True
            shape = (1, 1)

        dt = [(name, object) for name in fields]
        arr = np.empty(shape=shape, dtype=dt)

        for field in obj:
            obj_field = obj[field]
            for idx in np.ndindex(arr.shape):
                if is_scalar:
                    arr[idx][field] = self.read_h5_data(obj_field)
                else:
                    arr[idx][field] = self.read_h5_data(self.h5stream[obj_field[idx]])
        return arr.T

    def read_cell(self, obj, is_empty=0):
        """Reads MATLAB cell arrays from the v7.3 MAT-file."""

        if is_empty:
            return np.empty(shape=obj[()], dtype=object)

        arr = np.empty(shape=obj.shape, dtype=object)
        for idx in np.ndindex(obj.shape):
            ref_data = self.h5stream[obj[idx]]
            arr[idx] = self.read_h5_data(ref_data)
        return arr.T

    def read_sparse(self, obj, nrows):
        """Reads MATLAB sparse arrays from the v7.3 MAT-file."""

        jc = obj["jc"][()]
        ncols = jc.size - 1

        if "data" in obj:
            # Exists only if sparse matrix contains non-zero elements
            data = self.read_numeric(obj["data"])
            matlab_class = obj.attrs.get(MAT_HDF_ATTRS.CLASS, None)
            if data.dtype.kind != "c":
                dtype = matlab_class_to_dtype(matlab_class, data.dtype)
                data = data.astype(dtype)
            ir = obj["ir"][()]
        else:
            data = np.array([], dtype=np.float64)
            ir = np.array([], dtype=np.int32)

        return csc_array((data, ir, jc), shape=(nrows, ncols))

    def read_function_handle(self, obj, object_decode):
        """Reads MATLAB function handles from the v7.3 MAT-file."""

        if object_decode != ObjectDecodingHint.FUNCTION_HINT:
            raise ValueError(
                f"Function handle with object decode {object_decode} not supported. Data may be corrupted"
            )

        data = self.read_struct(obj)
        return MatlabFunction(data)

    def read_object(self, obj, classname):
        """Reads mxOBJECT_CLASS variables from the v7.3 MAT-file."""

        fields = self.read_struct(obj)
        return MatlabObject(fields, classname)

    def read_opaque(self, obj, object_decode):
        """Reads MATLAB opaque objects from the v7.3 MAT-file."""

        classname = obj.attrs.get(MAT_HDF_ATTRS.CLASS, None)
        if classname is not None:
            classname = classname.decode("ascii")

        if object_decode == ObjectDecodingHint.OBJECT_HINT:
            return self.read_object(obj, classname)

        if object_decode != ObjectDecodingHint.OPAQUE_HINT:
            raise ValueError(
                f"Opaque object with object decode {object_decode} not supported. Data may be corrupted"
            )

        if classname == MCOS_SUBSYSTEM_CLASS:
            return self.read_cell(obj)

        type_system = guess_type_system(classname)

        # Check Enumeration Instances
        fields = obj.attrs.get(MAT_HDF_ATTRS.FIELDS, None)
        if fields is not None:
            fields = ["".join(x.astype(str)) for x in fields]

            if ENUMERATION_INSTANCE_TAG in fields:
                metadata = self.read_struct(obj)
                if metadata[0, 0][ENUMERATION_INSTANCE_TAG] != MCOS_MAGIC_NUMBER:
                    return metadata
                return self.subsystem.load_mcos_enumeration(metadata, type_system)
            else:
                metadata = obj[()].T
        else:
            metadata = obj[()].T
        return self.subsystem.load_opaque_object(metadata, type_system, classname)

    def read_h5_data(self, obj):
        """Reads data from the HDF5 object."""

        matlab_class = obj.attrs.get(MAT_HDF_ATTRS.CLASS, None)
        is_empty = obj.attrs.get(MAT_HDF_ATTRS.EMPTY, 0)
        object_decode = obj.attrs.get(MAT_HDF_ATTRS.OBJECT_DECODE, -1)
        matlab_sparse = obj.attrs.get(MAT_HDF_ATTRS.SPARSE, -1)

        if matlab_sparse >= 0:
            arr = self.read_sparse(obj, matlab_sparse)
        elif matlab_class == b"char":
            arr = self.read_char(obj, is_empty)
        elif matlab_class in (
            b"int8",
            b"uint8",
            b"int16",
            b"uint16",
            b"int32",
            b"uint32",
            b"int64",
            b"uint64",
            b"single",
            b"double",
            b"logical",
        ):
            arr = self.read_numeric(obj, is_empty)
        elif matlab_class == b"struct":
            arr = self.read_struct(obj, is_empty)
        elif matlab_class == b"cell":
            arr = self.read_cell(obj, is_empty)
        elif matlab_class == b"function_handle":
            arr = self.read_function_handle(obj, object_decode)
        elif matlab_class == b"canonical empty" and is_empty:
            arr = MatlabCanonicalEmpty()
        elif object_decode >= 0:
            arr = self.read_opaque(obj, object_decode)
        else:
            raise NotImplementedError(f"MATLAB class {matlab_class} not supported")

        return arr

    def get_variables(self, variable_names):
        """Reads variables from the HDF5 file."""

        mdict = {}
        mdict["__globals__"] = []

        for var in self.h5stream:
            obj = self.h5stream[var]
            if var in (MAT_HDF_REFS_GROUP, MAT_HDF_SUBSYS_GROUP):
                continue
            if variable_names is not None and var not in variable_names:
                continue
            try:
                data = self.read_h5_data(obj)
            except Exception as err:
                raise ValueError(f"Error reading variable {var}: {err}") from err
            mdict[var] = data
            is_global = obj.attrs.get(MAT_HDF_ATTRS.GLOBAL, 0)
            if is_global:
                mdict["__globals__"].append(var)
            if variable_names is not None:
                variable_names.remove(var)
                if len(variable_names) == 0:
                    break

        return mdict


def read_subsystem(f, byte_order, raw_data, add_table_attrs):
    """Read subsystem data if present"""

    if MAT_HDF_SUBSYS_GROUP not in f:
        return None

    subsys_group = f[MAT_HDF_SUBSYS_GROUP]
    MR7 = MatRead7(f, None)
    subsys_data = MR7.read_struct(subsys_group)
    subsystem = MatSubsystem(byte_order, raw_data, add_table_attrs)
    subsystem.load_subsystem(subsys_data)

    return subsystem



import warnings
from enum import Enum

import numpy as np
import pandas as pd

from matio.utils.converters.matenum import (
    ENUM_INSTANCE_DTYPE,
    enum_to_opaque,
    mat_to_enum,
)
from matio.utils.converters.matmap import containermap_to_mat, mat_to_containermap
from matio.utils.converters.matstring import mat_to_string, string_to_mat
from matio.utils.converters.mattables import (
    categorical_to_mat,
    mat_to_categorical,
    mat_to_table,
    mat_to_timetable,
    table_to_mat,
    timetable_to_mat,
)
from matio.utils.converters.mattimes import (
    caldur_dtype,
    calendarduration_to_mat,
    datetime_to_mat,
    duration_to_mat,
    mat_to_calendarduration,
    mat_to_datetime,
    mat_to_duration,
)
from matio.utils.matclass import (
    MatlabContainerMap,
    MatlabEnumerationArray,
    MatlabOpaque,
    OpaqueType,
)

matlab_saveobj_ret_types = ["string", "timetable"]

matlab_classdef_types = [
    "calendarDuration",
    "categorical",
    "containers.Map",
    "datetime",
    # "dictionary",
    "duration",
    "string",
    "table",
    "timetable",
]

MAT_TO_PY = {
    "calendarDuration": mat_to_calendarduration,
    "categorical": mat_to_categorical,
    "containers.Map": mat_to_containermap,
    "datetime": mat_to_datetime,
    # "dictionary": mat_to_dictionary,
    "duration": mat_to_duration,
    "string": mat_to_string,
    "table": mat_to_table,
    "timetable": mat_to_timetable,
}

PY_TO_MAT = {
    "calendarDuration": calendarduration_to_mat,
    "categorical": categorical_to_mat,
    "containers.Map": containermap_to_mat,
    "datetime": datetime_to_mat,
    # "dictionary": dictionary_to_mat,
    "duration": duration_to_mat,
    "string": string_to_mat,
    "table": table_to_mat,
    "timetable": timetable_to_mat,
}


def convert_mat_to_py(props, classname, **kwargs):
    """Converts a MATLAB object to a Python object"""
    convert_func = MAT_TO_PY.get(classname)

    return convert_func(
        props,
        byte_order=kwargs.get("byte_order", None),
        add_table_attrs=kwargs.get("add_table_attrs", None),
    )


def guess_type_system(classname):
    """Gets the type system for the given class name."""

    # Possibly, MATLAB decodes internally based on some object attribute
    # This function is just guess work as of now
    # Used for decoding HDF5 files
    if classname.startswith("java.") or classname.startswith("com."):
        type_system = "java"
    elif classname.startswith("COM."):
        type_system = "handle"
    else:
        type_system = "MCOS"

    return type_system


def guess_class_name(data):
    """Guess the MATLAB class name for a given Python object"""

    if isinstance(data, pd.DataFrame):
        if pd.api.types.is_datetime64_any_dtype(
            data.index
        ) or pd.api.types.is_timedelta64_dtype(data.index):
            return "timetable"
        else:
            return "table"
    elif isinstance(data, pd.Categorical):
        return "categorical"
    elif isinstance(data, pd.Series):
        raise NotImplementedError("pandas.Series to MATLAB object not yet supported")
    elif isinstance(data, MatlabContainerMap):
        return "containers.Map"
    elif isinstance(data, (np.ndarray, np.generic)):
        if data.dtype.kind == "T":
            return "string"
        elif data.dtype.kind == "m":
            return "duration"
        elif data.dtype.kind == "M":
            return "datetime"
        elif data.dtype == caldur_dtype:
            return "calendarDuration"

    return None


def convert_py_to_mat(data, classname):
    """Convert a Python object to a MATLAB object"""

    convert_func = PY_TO_MAT.get(classname)
    type_system = guess_type_system(classname)
    prop_map = convert_func(data)

    obj = MatlabOpaque(
        properties=prop_map,
        classname=classname,
        type_system=type_system,
    )

    return obj



import os
import tempfile

import numpy as np
import pytest

from matio import load_from_mat, save_to_mat
from matio.utils.matclass import MatConvertWarning

files = [("test_time_v7.mat", "v7"), ("test_time_v73.mat", "v7.3")]


@pytest.mark.parametrize("filename, version", files)
class TestLoadMatlabDatetime:

    def test_datetime_scalar(self, filename, version):
        """Test reading datetime scalar data from MAT-file"""
        file_path = os.path.join(os.path.dirname(__file__), "data", filename)
        mdict = load_from_mat(file_path, variable_names=["dt_basic"])
        assert "dt_basic" in mdict

        dt_scalar = np.array([["2025-04-01T12:00:00"]], dtype="datetime64[ns]").reshape(
            1, 1
        )

        np.testing.assert_array_equal(mdict["dt_basic"], dt_scalar, strict=True)

    def test_datetime_array(self, filename, version):
        """Test reading datetime array data from MAT-file"""
        file_path = os.path.join(os.path.dirname(__file__), "data", filename)
        mdict = load_from_mat(file_path, variable_names=["dt_array"])
        assert "dt_array" in mdict

        dt_array = np.array(
            [
                [
                    "2025-04-01",
                    "2025-04-03",
                    "2025-04-05",
                    "2025-04-02",
                    "2025-04-04",
                    "2025-04-06",
                ]
            ],
            dtype="datetime64[ns]",
        ).reshape(2, 3)

        np.testing.assert_array_equal(mdict["dt_array"], dt_array, strict=True)

    def test_datetime_empty(self, filename, version):
        """Test reading empty datetime data from MAT-file"""
        file_path = os.path.join(os.path.dirname(__file__), "data", filename)
        mdict = load_from_mat(file_path, variable_names=["dt_empty"])
        assert "dt_empty" in mdict

        dt_empty = np.empty((0, 0), dtype="datetime64[ns]")
        np.testing.assert_array_equal(mdict["dt_empty"], dt_empty, strict=True)

    def test_datetime_fmt(self, filename, version):
        """Test reading datetime with format data from MAT-file"""
        file_path = os.path.join(os.path.dirname(__file__), "data", filename)
        with pytest.warns(MatConvertWarning, match="Ignoring 'fmt' property"):
            mdict = load_from_mat(file_path, variable_names=["dt_fmt"])
            assert "dt_fmt" in mdict

            dt_fmt = np.array(
                [["2025-04-01T12:00:00"]], dtype="datetime64[ns]"
            ).reshape(1, 1)

            np.testing.assert_array_equal(mdict["dt_fmt"], dt_fmt, strict=True)

    def test_datetime_tz(self, filename, version):
        """Test reading datetime with timezone data from MAT-file"""
        file_path = os.path.join(os.path.dirname(__file__), "data", filename)
        with pytest.warns(MatConvertWarning, match="does not support time zones"):
            mdict = load_from_mat(file_path, variable_names=["dt_tz"])
            assert "dt_tz" in mdict

            dt_tz = np.array([["2025-04-01T16:00:00"]], dtype="datetime64[ns]").reshape(
                1, 1
            )

            np.testing.assert_array_equal(mdict["dt_tz"], dt_tz, strict=True)


@pytest.mark.parametrize("filename, version", files)
class TestSaveMatlabDatetime:

    def test_datetime_scalar(self, filename, version):
        """Test writing datetime scalar to MAT-file"""
        file_path = os.path.join(os.path.dirname(__file__), "data", filename)
        mdict = load_from_mat(file_path, variable_names=["dt_basic"])

        with tempfile.NamedTemporaryFile(suffix=".mat", delete=False) as tmpfile:
            temp_file_path = tmpfile.name

        try:
            save_to_mat(temp_file_path, mdict, version=version)
            mload = load_from_mat(temp_file_path, variable_names=["dt_basic"])

            np.testing.assert_array_equal(
                mdict["dt_basic"], mload["dt_basic"], strict=True
            )

        finally:
            if os.path.exists(temp_file_path):
                os.remove(temp_file_path)

    def test_datetime_array(self, filename, version):
        """Test writing datetime array to MAT-file"""
        file_path = os.path.join(os.path.dirname(__file__), "data", filename)
        mdict = load_from_mat(file_path, variable_names=["dt_array"])

        with tempfile.NamedTemporaryFile(suffix=".mat", delete=False) as tmpfile:
            temp_file_path = tmpfile.name

        try:
            save_to_mat(temp_file_path, mdict, version=version)
            mload = load_from_mat(temp_file_path, variable_names=["dt_array"])

            np.testing.assert_array_equal(
                mdict["dt_array"], mload["dt_array"], strict=True
            )

        finally:
            if os.path.exists(temp_file_path):
                os.remove(temp_file_path)

    def test_datetime_empty(self, filename, version):
        """Test writing empty datetime to MAT-file"""
        file_path = os.path.join(os.path.dirname(__file__), "data", filename)
        mdict = load_from_mat(file_path, variable_names=["dt_empty"])

        with tempfile.NamedTemporaryFile(suffix=".mat", delete=False) as tmpfile:
            temp_file_path = tmpfile.name

        try:
            save_to_mat(temp_file_path, mdict, version=version)
            mload = load_from_mat(temp_file_path, variable_names=["dt_empty"])

            np.testing.assert_array_equal(
                mdict["dt_empty"], mload["dt_empty"], strict=True
            )

        finally:
            if os.path.exists(temp_file_path):
                os.remove(temp_file_path)



import os

import pytest

from matio import load_from_mat
from matio.utils.matclass import MatlabOpaque

files = ["test_figure.fig"]


@pytest.mark.parametrize("filename", files)
class TestLoadFig:

    def test_load_fig(self, filename):
        """Test reading fig file"""
        file_path = os.path.join(os.path.dirname(__file__), "data", filename)
        mdict = load_from_mat(file_path, variable_names=None)

        assert "hgS_070000" in mdict
        assert "hgM_070000" in mdict
        assert "meta_data" in mdict
        assert isinstance(mdict["hgM_070000"][0, 0]["GraphicsObjects"], MatlabOpaque)
        assert (
            mdict["hgM_070000"][0, 0]["GraphicsObjects"].classname
            == "matlab.graphics.internal.figfile.GraphicsObjects"
        )



import os
import tempfile

import numpy as np
import pytest

from matio import load_from_mat, save_to_mat
from matio.utils.matclass import MatlabOpaque

files = ["test_class_alias.mat"]


@pytest.mark.parametrize("filename", files)
class TestLoadClassAlias:

    def test_class_alias(self, filename):
        """Test reading class alias from MAT-file"""
        file_path = os.path.join(os.path.dirname(__file__), "data", filename)
        mdict = load_from_mat(file_path, variable_names=["obj1"])
        assert "obj1" in mdict
        assert isinstance(mdict["obj1"], MatlabOpaque)
        assert mdict["obj1"].class_alias == "SecondName"
        assert mdict["obj1"].classname == "FirstName"


@pytest.mark.parametrize("filename", files)
class TestSaveClassAlias:

    def test_class_alias(self, filename):
        """Test writing function handle simple to MAT-file"""
        file_path = os.path.join(os.path.dirname(__file__), "data", filename)
        mdict = load_from_mat(file_path, variable_names=["obj1"])

        with tempfile.NamedTemporaryFile(suffix=".mat", delete=False) as tmpfile:
            temp_file_path = tmpfile.name

        try:
            save_to_mat(temp_file_path, mdict)
            mload = load_from_mat(temp_file_path, variable_names=["obj1"])

            assert "obj1" in mload
            assert isinstance(mload["obj1"], MatlabOpaque)
            assert mload["obj1"].class_alias == mdict["obj1"].class_alias
            assert mload["obj1"].class_alias == "SecondName"
            assert mload["obj1"].classname == mdict["obj1"].classname

        finally:
            if os.path.exists(temp_file_path):
                os.remove(temp_file_path)



import os
import tempfile

import numpy as np
import pandas as pd
import pytest

from matio import load_from_mat, save_to_mat
from matio.utils.matclass import MatlabOpaque

files = [("test_tables_v7.mat", "v7"), ("test_tables_v73.mat", "v7.3")]
namespace = "TestClasses"


@pytest.mark.parametrize("filename, version", files)
class TestLoadMatlabTable:

    def test_table_numeric(self, filename, version):
        """Test reading datetime scalar data from MAT-file"""
        file_path = os.path.join(os.path.dirname(__file__), "data", filename)
        mdict = load_from_mat(file_path, variable_names=["table_numeric"])
        assert "table_numeric" in mdict

        df = pd.DataFrame({"Var1": [1.1, 2.2, 3.3], "Var2": [4.4, 5.5, 6.6]})
        pd.testing.assert_frame_equal(mdict["table_numeric"], df, check_like=True)

    def test_table_strings(self, filename, version):
        """Test reading table with string data from MAT-file"""
        file_path = os.path.join(os.path.dirname(__file__), "data", filename)
        mdict = load_from_mat(file_path, variable_names=["table_strings"])
        assert "table_strings" in mdict

        df = pd.DataFrame({"Var1": ["apple", "banana", "cherry"]})
        pd.testing.assert_frame_equal(mdict["table_strings"], df, check_like=True)

    def test_table_empty(self, filename, version):
        """Test reading empty table from MAT-file"""
        file_path = os.path.join(os.path.dirname(__file__), "data", filename)
        mdict = load_from_mat(file_path, variable_names=["table_empty"])
        assert "table_empty" in mdict

        df = pd.DataFrame()
        pd.testing.assert_frame_equal(mdict["table_empty"], df, check_like=True)

    def test_table_time(self, filename, version):
        """Test reading timetable from MAT-file"""
        file_path = os.path.join(os.path.dirname(__file__), "data", filename)
        mdict = load_from_mat(file_path, variable_names=["table_time"])
        assert "table_time" in mdict

        df = pd.DataFrame(
            {
                "Time": np.array(
                    [
                        "2020-01-01T00:00:00.000",
                        "2020-01-02T00:00:00.000",
                        "2020-01-03T00:00:00.000",
                    ],
                    dtype="datetime64[ns]",
                ),
                "Duration": np.array([30, 60, 90], dtype="timedelta64[s]"),
            }
        )
        pd.testing.assert_frame_equal(mdict["table_time"], df, check_like=True)

    def test_table_nan(self, filename, version):
        """Test reading table with NaN values from MAT-file"""
        file_path = os.path.join(os.path.dirname(__file__), "data", filename)
        mdict = load_from_mat(file_path, variable_names=["table_nan"])
        assert "table_nan" in mdict

        df = pd.DataFrame(
            {
                "Var1": [1.1, np.nan, 3.3],
                "Var2": np.array(["A", "", "C"]),
            }
        )
        pd.testing.assert_frame_equal(mdict["table_nan"], df, check_like=True)

    def test_table_from_cell(self, filename, version):
        """Test reading table created from cell array in MAT-file"""
        file_path = os.path.join(os.path.dirname(__file__), "data", filename)
        mdict = load_from_mat(file_path, variable_names=["table_from_cell"])
        assert "table_from_cell" in mdict

        df = pd.DataFrame(
            {
                "Var1": [
                    np.array([[1.0]]),
                    np.array(["text"]),
                    np.array([["2023-01-01T00:00:00.000"]], dtype="datetime64[ns]"),
                ],
            }
        )
        pd.testing.assert_frame_equal(mdict["table_from_cell"], df, check_like=True)

    def test_table_var_names(self, filename, version):
        """Test reading table with variable names from MAT-file"""
        file_path = os.path.join(os.path.dirname(__file__), "data", filename)
        mdict = load_from_mat(file_path, variable_names=["table_var_names"])
        assert "table_var_names" in mdict

        df = pd.DataFrame(
            {
                "X": np.array([10.0, 20.0, 30.0]),
                "Y": np.array([100.0, 200.0, 300.0]),
            }
        )
        pd.testing.assert_frame_equal(mdict["table_var_names"], df, check_like=True)

    def test_table_row_names(self, filename, version):
        """Test reading table with row names from MAT-file"""
        file_path = os.path.join(os.path.dirname(__file__), "data", filename)
        mdict = load_from_mat(file_path, variable_names=["table_row_names"])
        assert "table_row_names" in mdict

        df = pd.DataFrame(
            {
                "Col1": np.array([1.0, 2.0, 3.0]),
                "Col2": np.array([4.0, 5.0, 6.0]),
            },
            index=["R1", "R2", "R3"],
        )
        pd.testing.assert_frame_equal(mdict["table_row_names"], df, check_like=True)

    def test_table_multi_col_data(self, filename, version):
        """Test reading table with multi-column data from MAT-file"""
        file_path = os.path.join(os.path.dirname(__file__), "data", filename)
        mdict = load_from_mat(file_path, variable_names=["table_multi_col_data"])
        assert "table_multi_col_data" in mdict

        df = pd.DataFrame(
            {
                "Var1": np.array(
                    [
                        "2023-01-01T00:00:00.000",
                        "2023-01-02T00:00:00.000",
                        "2023-01-03T00:00:00.000",
                    ],
                    dtype="datetime64[ns]",
                ),
                "multicoldata_1": np.array([1.0, 2.0, 3.0]),
                "multicoldata_2": np.array([4.0, 5.0, 6.0]),
            }
        )
        pd.testing.assert_frame_equal(
            mdict["table_multi_col_data"], df, check_like=True
        )

    def test_table_with_objects(self, filename, version):
        """Test reading table with object data from MAT-file"""
        file_path = os.path.join(os.path.dirname(__file__), "data", filename)
        mdict = load_from_mat(file_path, variable_names=["table_with_objects"])
        assert "table_with_objects" in mdict

        struct_field1 = np.array([[123]], dtype=np.float64)
        struct_field2 = np.array(["abc"], dtype=np.str_)

        col1 = mdict["table_with_objects"]["C"]
        for i in range(col1.size):
            assert isinstance(col1[i], np.ndarray)
            assert col1[i].dtype.hasobject
            assert set(col1[i].dtype.names) == {"field1", "field2"}
            np.testing.assert_array_equal(col1[i]["field1"][0, 0], struct_field1)
            np.testing.assert_array_equal(col1[i]["field2"][0, 0], struct_field2)

        col2 = mdict["table_with_objects"]["Var2"]
        for i in range(col2.size):
            assert isinstance(col2[i], MatlabOpaque)
            assert col2[i].classname == f"{namespace}.BasicClass"
            assert col2[i].type_system == "MCOS"
            np.testing.assert_array_equal(
                col2[i].properties["a"], np.array([[1]], dtype=np.float64)
            )

    def test_table_with_attrs(self, filename, version):
        """Test reading table with attributes from MAT-file"""
        file_path = os.path.join(os.path.dirname(__file__), "data", filename)
        mdict = load_from_mat(
            file_path, variable_names=["table_with_attrs"], add_table_attrs=True
        )
        assert "table_with_attrs" in mdict

        df = pd.DataFrame(
            {
                "ID": np.array([1.0, 2.0]),
                "Label": np.array(["one", "two"]),
            }
        )
        pd.testing.assert_frame_equal(mdict["table_with_attrs"], df, check_like=True)

        attrs = {
            "Description": "Test table with full metadata",
            "DimensionNames": ["RowId", "Features"],
            "VariableUnits": ["", "category"],
            "VariableDescriptions": ["ID number", "Category label"],
            "VariableContinuity": ["continuous", "step"],
            "UserData": np.array(
                [[(np.array(["UnitTest"]), np.array([[1.0]]))]],
                dtype=[("CreatedBy", "O"), ("Version", "O")],
            ),
        }

        for key, value in attrs.items():
            assert key in mdict["table_with_attrs"].attrs
            if isinstance(value, np.ndarray):
                np.testing.assert_array_equal(
                    mdict["table_with_attrs"].attrs[key], value
                )
            else:
                assert mdict["table_with_attrs"].attrs[key] == value


@pytest.mark.parametrize("filename, version", files)
class TestSaveMatlabTable:

    def test_table_numeric(self, filename, version):
        """Test writing numeric table to MAT-file"""
        file_path = os.path.join(os.path.dirname(__file__), "data", filename)
        mdict = load_from_mat(file_path, variable_names=["table_numeric"])

        with tempfile.NamedTemporaryFile(suffix=".mat", delete=False) as tmpfile:
            temp_file_path = tmpfile.name

        try:
            save_to_mat(temp_file_path, mdict, version=version)
            mload = load_from_mat(temp_file_path, variable_names=["table_numeric"])

            pd.testing.assert_frame_equal(
                mdict["table_numeric"], mload["table_numeric"], check_like=True
            )

        finally:
            if os.path.exists(temp_file_path):
                os.remove(temp_file_path)

    def test_table_strings(self, filename, version):
        """Test writing string table to MAT-file"""
        file_path = os.path.join(os.path.dirname(__file__), "data", filename)
        mdict = load_from_mat(file_path, variable_names=["table_strings"])

        with tempfile.NamedTemporaryFile(suffix=".mat", delete=False) as tmpfile:
            temp_file_path = tmpfile.name

        try:
            save_to_mat(temp_file_path, mdict, version=version)
            mload = load_from_mat(temp_file_path, variable_names=["table_strings"])

            pd.testing.assert_frame_equal(
                mdict["table_strings"], mload["table_strings"], check_like=True
            )

        finally:
            if os.path.exists(temp_file_path):
                os.remove(temp_file_path)

    def test_table_empty(self, filename, version):
        """Test writing empty table to MAT-file"""
        file_path = os.path.join(os.path.dirname(__file__), "data", filename)
        mdict = load_from_mat(file_path, variable_names=["table_empty"])

        with tempfile.NamedTemporaryFile(suffix=".mat", delete=False) as tmpfile:
            temp_file_path = tmpfile.name

        try:
            save_to_mat(temp_file_path, mdict, version=version)
            mload = load_from_mat(temp_file_path, variable_names=["table_empty"])

            pd.testing.assert_frame_equal(
                mdict["table_empty"], mload["table_empty"], check_like=True
            )

        finally:
            if os.path.exists(temp_file_path):
                os.remove(temp_file_path)

    def test_table_time(self, filename, version):
        """Test writing timetable to MAT-file"""
        file_path = os.path.join(os.path.dirname(__file__), "data", filename)
        mdict = load_from_mat(file_path, variable_names=["table_time"])

        with tempfile.NamedTemporaryFile(suffix=".mat", delete=False) as tmpfile:
            temp_file_path = tmpfile.name

        try:
            save_to_mat(temp_file_path, mdict, version=version)
            mload = load_from_mat(temp_file_path, variable_names=["table_time"])

            pd.testing.assert_frame_equal(
                mdict["table_time"], mload["table_time"], check_like=True
            )

        finally:
            if os.path.exists(temp_file_path):
                os.remove(temp_file_path)

    def test_table_nan(self, filename, version):
        """Test writing table with NaN values to MAT-file"""
        file_path = os.path.join(os.path.dirname(__file__), "data", filename)
        mdict = load_from_mat(file_path, variable_names=["table_nan"])

        with tempfile.NamedTemporaryFile(suffix=".mat", delete=False) as tmpfile:
            temp_file_path = tmpfile.name

        try:
            save_to_mat(temp_file_path, mdict, version=version)
            mload = load_from_mat(temp_file_path, variable_names=["table_nan"])

            pd.testing.assert_frame_equal(
                mdict["table_nan"], mload["table_nan"], check_like=True
            )

        finally:
            if os.path.exists(temp_file_path):
                os.remove(temp_file_path)

    def test_table_from_cell(self, filename, version):
        """Test writing table created from cell array to MAT-file"""
        file_path = os.path.join(os.path.dirname(__file__), "data", filename)
        mdict = load_from_mat(file_path, variable_names=["table_from_cell"])

        with tempfile.NamedTemporaryFile(suffix=".mat", delete=False) as tmpfile:
            temp_file_path = tmpfile.name

        try:
            save_to_mat(temp_file_path, mdict, version=version)
            mload = load_from_mat(temp_file_path, variable_names=["table_from_cell"])

            pd.testing.assert_frame_equal(
                mdict["table_from_cell"], mload["table_from_cell"], check_like=True
            )

        finally:
            if os.path.exists(temp_file_path):
                os.remove(temp_file_path)

    def test_table_var_names(self, filename, version):
        """Test writing table with variable names to MAT-file"""
        file_path = os.path.join(os.path.dirname(__file__), "data", filename)
        mdict = load_from_mat(file_path, variable_names=["table_var_names"])

        with tempfile.NamedTemporaryFile(suffix=".mat", delete=False) as tmpfile:
            temp_file_path = tmpfile.name

        try:
            save_to_mat(temp_file_path, mdict, version=version)
            mload = load_from_mat(temp_file_path, variable_names=["table_var_names"])

            pd.testing.assert_frame_equal(
                mdict["table_var_names"], mload["table_var_names"], check_like=True
            )

        finally:
            if os.path.exists(temp_file_path):
                os.remove(temp_file_path)

    def test_table_row_names(self, filename, version):
        """Test writing table with row names to MAT-file"""
        file_path = os.path.join(os.path.dirname(__file__), "data", filename)
        mdict = load_from_mat(file_path, variable_names=["table_row_names"])

        with tempfile.NamedTemporaryFile(suffix=".mat", delete=False) as tmpfile:
            temp_file_path = tmpfile.name

        try:
            save_to_mat(temp_file_path, mdict, version=version)
            mload = load_from_mat(temp_file_path, variable_names=["table_row_names"])

            pd.testing.assert_frame_equal(
                mdict["table_row_names"], mload["table_row_names"], check_like=True
            )

        finally:
            if os.path.exists(temp_file_path):
                os.remove(temp_file_path)

    def test_table_multi_col_data(self, filename, version):
        """Test writing table with multi-column data to MAT-file"""
        file_path = os.path.join(os.path.dirname(__file__), "data", filename)
        mdict = load_from_mat(file_path, variable_names=["table_multi_col_data"])

        with tempfile.NamedTemporaryFile(suffix=".mat", delete=False) as tmpfile:
            temp_file_path = tmpfile.name

        try:
            save_to_mat(temp_file_path, mdict, version=version)
            mload = load_from_mat(
                temp_file_path, variable_names=["table_multi_col_data"]
            )

            pd.testing.assert_frame_equal(
                mdict["table_multi_col_data"],
                mload["table_multi_col_data"],
                check_like=True,
            )

        finally:
            if os.path.exists(temp_file_path):
                os.remove(temp_file_path)

    def test_table_with_objects(self, filename, version):
        """Test writing table with object data to MAT-file"""
        file_path = os.path.join(os.path.dirname(__file__), "data", filename)
        mdict = load_from_mat(file_path, variable_names=["table_with_objects"])

        with tempfile.NamedTemporaryFile(suffix=".mat", delete=False) as tmpfile:
            temp_file_path = tmpfile.name

        try:
            save_to_mat(temp_file_path, mdict, version=version)
            mload = load_from_mat(temp_file_path, variable_names=["table_with_objects"])

            col1_orig = mdict["table_with_objects"]["C"]
            col1_load = mload["table_with_objects"]["C"]
            for i in range(col1_orig.size):
                assert isinstance(col1_load[i], np.ndarray)
                assert col1_load[i].dtype.hasobject
                assert set(col1_load[i].dtype.names) == {"field1", "field2"}
                np.testing.assert_array_equal(
                    col1_load[i]["field1"][0, 0], col1_orig[i]["field1"][0, 0]
                )
                np.testing.assert_array_equal(
                    col1_load[i]["field2"][0, 0], col1_orig[i]["field2"][0, 0]
                )

            col2_orig = mdict["table_with_objects"]["Var2"]
            col2_load = mload["table_with_objects"]["Var2"]
            for i in range(col2_orig.size):
                assert isinstance(col2_load[i], MatlabOpaque)
                assert col2_load[i].classname == col2_orig[i].classname
                assert col2_load[i].type_system == col2_orig[i].type_system
                np.testing.assert_array_equal(
                    col2_load[i].properties["a"], col2_orig[i].properties["a"]
                )

        finally:
            if os.path.exists(temp_file_path):
                os.remove(temp_file_path)

    @pytest.mark.xfail(reason="Writing attributes not implemented")
    def test_table_with_attrs(self, filename, version):
        """Test writing table with attributes to MAT-file"""
        file_path = os.path.join(os.path.dirname(__file__), "data", filename)
        mdict = load_from_mat(
            file_path, variable_names=["table_with_attrs"], add_table_attrs=True
        )

        with tempfile.NamedTemporaryFile(suffix=".mat", delete=False) as tmpfile:
            temp_file_path = tmpfile.name

        try:
            save_to_mat(temp_file_path, mdict, version=version, do_compression=False)
            mload = load_from_mat(
                temp_file_path,
                variable_names=["table_with_attrs"],
                add_table_attrs=True,
            )

            pd.testing.assert_frame_equal(
                mdict["table_with_attrs"],
                mload["table_with_attrs"],
                check_like=True,
            )

            for key in mdict["table_with_attrs"].attrs.keys():
                assert key in mload["table_with_attrs"].attrs
                val_orig = mdict["table_with_attrs"].attrs[key]
                val_load = mload["table_with_attrs"].attrs[key]
                if isinstance(val_orig, np.ndarray):
                    np.testing.assert_array_equal(val_orig, val_load)
                else:
                    assert val_orig == val_load

        finally:
            if os.path.exists(temp_file_path):
                os.remove(temp_file_path)



"""Reader for MAT-5 version files"""

# Copyright (c) 2001-2002 Enthought, Inc. 2003, SciPy Developers.
# All rights reserved.
#
# Modified by foreverallama (c) 2025
# https://github.com/foreverallama/matio
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions
# are met:
#
# 1. Redistributions of source code must retain the above copyright
#    notice, this list of conditions and the following disclaimer.
#
# 2. Redistributions in binary form must reproduce the above
#    copyright notice, this list of conditions and the following
#    disclaimer in the documentation and/or other materials provided
#    with the distribution.
#
# 3. Neither the name of the copyright holder nor the names of its
#    contributors may be used to endorse or promote products derived
#    from this software without specific prior written permission.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
# "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
# LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR
# A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT
# OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
# SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
# LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
# DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
# THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
# (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
# OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

import sys
import warnings
from io import BytesIO

from matio.subsystem import MatSubsystem
from matio.utils.matclass import MatReadError, MatReadWarning
from matio.utils.matheaders import MAT5_HEADER_SIZE_BYTES, check_mat_version
from matio.utils.matutils import shape_from_metadata

# Constants and helper objects
from matio.v5.matio5_params import mclass_info, miTypes, mxTypes

# Reader object for matlab 5 format variables
from ._mio5_utils import VarReader5
from ._streams import ZlibInputStream

SYS_BYTE_ORDER = "<" if sys.byteorder == "little" else ">"


def loadmat5(
    file_path, subsystem_offset, byte_order, variable_names, raw_data, add_table_attrs
):
    """Load MATLAB file."""

    with open(file_path, "rb") as f:

        subsystem = read_subsystem(
            f,
            subsystem_offset,
            byte_order,
            raw_data=raw_data,
            add_table_attrs=add_table_attrs,
        )

        MR = MatFile5Reader(f, byte_order=byte_order, subsystem=subsystem)
        matfile_dict = MR.get_variables(variable_names)

    return matfile_dict


def whosmat5(file_path, byte_order):
    """List variables in MATLAB file."""

    with open(file_path, "rb") as f:
        MR = MatFile5Reader(f, byte_order=byte_order)
        vars = MR.list_variables()

    return vars


class MatFile5Reader:
    """Reader for Mat 5 mat files"""

    def __init__(
        self,
        mat_stream,
        byte_order=SYS_BYTE_ORDER,
        subsystem=None,
    ):
        """Initialize reader for Mat-5 files"""

        self.mat_stream = mat_stream
        self.dtypes = {}

        self.byte_order = byte_order
        self.uint16_codec = sys.getdefaultencoding()  # scipy arg

        self._file_reader = None
        self._matrix_reader = None
        self.subsystem = subsystem

    def end_of_stream(self):
        curpos = self.mat_stream.tell()
        self.mat_stream.seek(0, 2)
        endpos = self.mat_stream.tell()
        self.mat_stream.seek(curpos)
        return curpos == endpos

    def initialize_read(self):
        """Initialize file-level and matrix-level readers"""
        # reader for top level stream. We need this extra top-level
        # reader because we use the matrix_reader object to contain
        # compressed matrices (so they have their own stream)
        self._file_reader = VarReader5(self)
        self._matrix_reader = VarReader5(self)

    def read_var_header(self):
        """Read variable headers"""

        mdtype, byte_count = self._file_reader.read_full_tag()
        if not byte_count > 0:
            raise MatReadError("Did not read any bytes")
        next_pos = self.mat_stream.tell() + byte_count
        if mdtype == miTypes.miCOMPRESSED:
            # Make new stream from compressed data
            stream = ZlibInputStream(self.mat_stream, byte_count)
            self._matrix_reader.set_stream(stream)
            check_stream_limit = True
            mdtype, byte_count = self._matrix_reader.read_full_tag()
        else:
            check_stream_limit = False
            self._matrix_reader.set_stream(self.mat_stream)
        if not mdtype == miTypes.miMATRIX:
            raise TypeError(f"Expecting miMATRIX type here, got {mdtype}")
        header = self._matrix_reader.read_header(check_stream_limit)
        return header, next_pos

    def read_var_array(self, header, process=True):
        """Read variable array from stream"""
        return self._matrix_reader.array_from_header(header, process)

    def get_variables(self, variable_names=None):
        """Get variables from stream"""

        self.mat_stream.seek(MAT5_HEADER_SIZE_BYTES)
        self.initialize_read()

        mdict = {"__globals__": []}

        while not self.end_of_stream():
            hdr, next_position = self.read_var_header()
            name = hdr.name.decode("ascii")
            if name in mdict:
                msg = f"Duplicate variable name {name!r} in file. Overwriting previous."
                warnings.warn(msg, MatReadWarning, stacklevel=2)
            if name == "":
                # Skip subsystem
                self.mat_stream.seek(next_position)
                continue
            if variable_names is not None and name not in variable_names:
                self.mat_stream.seek(next_position)
                continue
            try:
                res = self.read_var_array(hdr, True)
            except MatReadError as err:
                warnings.warn(
                    f'Unreadable variable "{name}", because "{err}"',
                    Warning,
                    stacklevel=2,
                )
                res = f"Read error: {err}"
            self.mat_stream.seek(next_position)
            mdict[name] = res
            if hdr.is_global:
                mdict["__globals__"].append(name)
            if variable_names is not None:
                variable_names.remove(name)
                if len(variable_names) == 0:
                    break

        return mdict

    def list_variables(self):
        """List variables from stream"""
        self.mat_stream.seek(MAT5_HEADER_SIZE_BYTES)
        self.initialize_read()
        vars = []
        while not self.end_of_stream():
            hdr, next_position = self.read_var_header()
            name = hdr.name.decode("ascii")
            if name == "":
                # Skip subsystem
                self.mat_stream.seek(next_position)
                continue
            if hdr.mclass == mxTypes.mxOPAQUE_CLASS:
                shape = self.read_opaque_class_shape(hdr)
            else:
                shape = self._matrix_reader.shape_from_header(hdr)

            if hdr.is_logical:
                info = "logical"
            elif hdr.classname is not None:
                info = hdr.classname
            else:
                info = mclass_info.get(hdr.mclass, "unknown")

            vars.append((name, shape, info))

            self.mat_stream.seek(next_position)
        return vars

    def read_opaque_class_shape(self, hdr):
        """Read class info for OBJECT and OPAQUE types"""

        objmetadata = self._matrix_reader.array_from_header(hdr, False)
        shape = shape_from_metadata(objmetadata)
        return shape


def load_subsys_stream(f, subsystem_offset, byte_order):
    """Load subsystem stream if present"""

    f.seek(subsystem_offset)
    subsys_stream_reader = MatFile5Reader(f, byte_order=byte_order)
    subsys_stream_reader.initialize_read()
    hdr, _ = subsys_stream_reader.read_var_header()

    try:
        res = subsys_stream_reader.read_var_array(hdr, False)
    except MatReadError as err:
        raise MatReadError(f'Unreadable subsystem data because "{err}"')

    # Subsystem data is a file stream written as uint8 data
    # FIXME: Can this be optimized?
    return BytesIO(res)


def read_subsystem(f, subsystem_offset, byte_order, raw_data, add_table_attrs):
    """Read subsystem data if present"""

    if subsystem_offset <= 0:
        return None

    subsys_stream = load_subsys_stream(f, subsystem_offset, byte_order)
    subsys_stream.seek(0)
    byte_order, _ = check_mat_version(subsys_stream.read(4))
    subsys_stream.read(4)  # Discard Padding

    subsystem_reader = MatFile5Reader(subsys_stream, byte_order=byte_order)
    subsystem_reader.initialize_read()

    hdr, _ = subsystem_reader.read_var_header()
    try:
        subsys_data = subsystem_reader.read_var_array(hdr, False)
    except MatReadError as err:
        raise MatReadError(f'Unreadable subsystem data because "{err}"')

    subsystem = MatSubsystem(
        byte_order, raw_data=raw_data, add_table_attrs=add_table_attrs
    )
    subsystem.load_subsystem(subsys_data)

    return subsystem



from enum import Enum

import numpy as np

from matio.utils.matclass import MatlabEnumerationArray, MatlabOpaque, OpaqueType
from matio.utils.matheaders import MCOS_MAGIC_NUMBER

ENUM_INSTANCE_DTYPE = [
    ("EnumerationInstanceTag", object),
    ("ClassName", object),
    ("ValueNames", object),
    ("Values", object),
    ("ValueIndices", object),
    ("BuiltinClassName", object),
]


def mat_to_enum(values, value_names, classname, shapes):
    """Converts MATLAB enum to Python enum"""

    enum_class = Enum(
        classname,
        {name: val.properties for name, val in zip(value_names, values)},
    )

    enum_members = [enum_class(val.properties) for val in values]
    arr = np.array(enum_members, dtype=object).reshape(shapes, order="F")
    return MatlabEnumerationArray(arr, type_system=OpaqueType.MCOS, classname=classname)


def enum_to_opaque(classname, props):
    """Converts Python enum to MATLAB enum"""

    return MatlabOpaque(
        properties=props, classname=classname, type_system=OpaqueType.MCOS
    )



"""Utility functions for matio"""

import warnings
from collections.abc import Mapping

import numpy as np
from scipy.sparse import issparse

from matio.utils.matclass import (
    EmptyMatStruct,
    IntegerDecodingHint,
    MatlabCanonicalEmpty,
    MatlabClasses,
    MatlabOpaque,
    MatWriteWarning,
)
from matio.utils.matconvert import convert_py_to_mat, guess_class_name
from matio.utils.matheaders import MAT_5_VERSION, MAT_HDF_VERSION


def chars_to_strings(in_arr):
    """Convert a numpy array of characters to an array of strings.
    Notes:
    Scipy basically just collapses the last axis into the string length.
    """
    arr = in_arr
    ndim = arr.ndim
    last_dim = arr.shape[-1]

    # Handle empty last axis
    if last_dim == 0:
        new_dt_str = arr.dtype.str
        if ndim == 2:
            out_shape = (0,)
        else:
            out_shape = arr.shape[:-2] + (0,)
    else:
        # Construct new dtype with last_dim as string length
        new_dt_str = arr.dtype.str[:-1] + str(last_dim)
        out_shape = arr.shape[:-1]

    arr = np.ascontiguousarray(arr)
    arr = arr.view(new_dt_str)
    return arr.reshape(out_shape)


def _get_string_arr_dtype(arr, num=1):
    """Return dtype for given number of items per element"""
    if arr.dtype.kind == "U":
        return np.dtype(f"U{num}")
    elif arr.dtype.kind == "S":
        return np.dtype(f"S{num}")


def strings_to_chars(arr):
    """Convert numpy string array to char array.
    Notes:
    Expands 1D numpy string arrays to char arrays.
    Basically inverse of chars_to_strings in scipy.

    2D numpy string arrays will be treated as MATLAB strings instead
    """
    dims = list(arr.shape)
    if not dims:
        dims = [1]
    dims.append(int(arr.dtype.str[2:]))
    arr = np.ndarray(shape=dims, dtype=_get_string_arr_dtype(arr), buffer=arr)
    empties = [arr == np.array("", dtype=arr.dtype)]
    if not np.any(empties):
        return arr
    arr = arr.copy()
    arr[tuple(empties)] = " "
    return arr


def matlab_class_to_dtype(matlab_class, current_dtype):
    """Map MATLAB class names to numpy dtypes.
    Used for v7.3 format
    """
    matlab_class = matlab_class.decode("ascii")
    try:
        dtype = np.dtype(matlab_class)
    except TypeError:
        if matlab_class == MatlabClasses.LOGICAL:
            dtype = np.dtype(np.bool_)
        else:
            # Fallback
            dtype = current_dtype
    return dtype


def matdims(arr, oned_as="col"):
    """Determine equivalent MATLAB dimensions for given array"""
    shape = arr.shape
    if shape == ():  # scalar
        return (1, 1)
    if len(shape) == 1:  # 1D
        if shape[0] == 0:
            return (0, 0)
        elif oned_as == "col":
            return shape + (1,)
        elif oned_as == "row":
            return (1,) + shape
        else:
            raise ValueError(f'Expected oned_as to be "col" or "row", got {oned_as!r}')
    return shape


def mat_numeric(arr, version, classname=None, int_decode=None):
    """Convert numpy numeric type to MATLAB compatible type"""

    dt = arr.dtype
    if dt.kind in ("i", "u"):
        if dt.itemsize <= 8:
            target_dtype = dt
            classname = target_dtype.name
        else:
            target_dtype = np.int64 if dt.kind == "i" else np.uint64
            target_dtype = np.dtype(target_dtype)
            classname = target_dtype.name
            warnings.warn(
                f"Integer type {dt} not supported in MATLAB. Converting to {target_dtype}.",
                MatWriteWarning,
            )
    elif dt.kind == "f":
        if dt.itemsize == 4:
            target_dtype = np.dtype(np.float32)
            classname = MatlabClasses.SINGLE
        elif dt.itemsize == 8:
            target_dtype = np.dtype(np.float64)
            classname = MatlabClasses.DOUBLE
        else:
            target_dtype = np.dtype(np.float64)
            classname = MatlabClasses.DOUBLE
            warnings.warn(
                f"Float type {dt} not supported in MATLAB. Converting to {target_dtype}.",
                MatWriteWarning,
            )
    elif dt.kind == "c":
        if version == MAT_5_VERSION:
            target_dtype = np.dtype(np.complex128)
            if dt.itemsize != 16:
                warnings.warn(
                    f"Complex type {dt} not supported in MATLAB. Converting to {target_dtype}.",
                    MatWriteWarning,
                )
        elif version == MAT_HDF_VERSION:
            target_dtype = np.dtype([("real", np.float64), ("imag", np.float64)])
            classname = "double"
            if dt.itemsize != 16:
                warnings.warn(
                    f"Complex type {dt} not supported in MATLAB. Converting to {target_dtype}.",
                    MatWriteWarning,
                )

            data_new = np.empty(arr.shape, dtype=target_dtype)
            data_new["real"] = arr.real
            data_new["imag"] = arr.imag
            arr = data_new
        else:
            raise ValueError(f"Unknown MAT-file version '{version}' specified")

    elif dt.kind == "b":
        target_dtype = np.dtype(np.uint8)
        int_decode = IntegerDecodingHint.LOGICAL_HINT
        classname = MatlabClasses.LOGICAL

    # arr = arr.astype(target_dtype, copy=True)
    arr = arr.astype(target_dtype)
    return arr, classname, int_decode


def to_writeable(source, oned_as="col"):
    """Convert input object ``source`` to something we can write

    Parameters
    ----------
    source : object

    Returns
    -------
    arr : None or ndarray or EmptyStructMarker
        If `source` cannot be converted to something we can write to a matfile,
        return None.  If `source` is equivalent to an empty dictionary, return
        ``EmptyStructMarker``.  Otherwise return `source` converted to an
        ndarray with contents for writing to matfile.
    """
    if source is None:
        return np.empty((0, 0), dtype=np.float64)

    if isinstance(source, (MatlabOpaque, MatlabCanonicalEmpty)):
        return source

    if issparse(source):
        return source

    classname = guess_class_name(source)
    if classname is not None:
        if isinstance(source, (np.ndarray, np.generic)):
            source = np.asanyarray(source)
            source = np.atleast_2d(source)
            if source.ndim == 1 and oned_as == "col":
                source = source.T
        return convert_py_to_mat(source, classname)

    if isinstance(source, (np.ndarray, np.generic)):
        narr = np.asanyarray(source)

        if narr.dtype.kind in ("U", "S"):
            # Backwards compatibility with scipy "chars_to_strings"
            return narr

        narr = np.atleast_2d(narr)
        if narr.ndim == 1 and oned_as == "col":
            narr = narr.T

        return narr

    if hasattr(source, "__array__"):
        return np.asarray(source)

    if isinstance(source, Mapping):
        is_mapping = True
    elif hasattr(source, "__dict__"):
        source = {
            key: value
            for key, value in source.__dict__.items()
            if not key.startswith("_")
        }
        is_mapping = True
    else:
        is_mapping = False

    if is_mapping:
        dtype = []
        values = []
        for field, value in source.items():
            if isinstance(field, str):
                if field[0] not in "_0123456789":
                    dtype.append((str(field), object))
                    values.append(value)
                else:
                    msg = (
                        f"Starting field name with a underscore "
                        f"or a digit ({field}) is ignored"
                    )
                    warnings.warn(msg, MatWriteWarning, stacklevel=2)
        if dtype:
            return np.array([tuple(values)], dtype)
        else:
            return EmptyMatStruct(np.array([]))

    # Try and convert to numpy array
    try:
        narr = np.asanyarray(source)
    except ValueError:
        narr = np.asanyarray(source, dtype=object)

    if narr.dtype.type in (object, np.object_) and narr.shape == () and narr == source:
        raise TypeError(f"Could not convert {type(source)} to a writeable datatype")

    if narr.dtype.kind in ("U", "S"):
        # Backwards compatibility with scipy "chars_to_strings"
        return np.atleast_1d(narr)

    narr = np.atleast_2d(narr)
    if oned_as == "col":
        narr = narr.T
    return narr


def shape_from_metadata(metadata):
    """Extract shape from MATLAB object metadata"""

    if not isinstance(metadata, np.ndarray):
        return ()

    if metadata.dtype == np.uint32:
        ndims = metadata.flat[1]
        dims = [int(x) for x in metadata.flat[2 : 2 + ndims]]

    elif metadata.dtype.hasobject and metadata.dtype.names:
        # Enumeration Array
        dims = metadata[0, 0]["ValueIndices"].shape

    else:
        dims = metadata.shape

    return tuple(dims)


def sanitize_input_lists(var_list, arg_name):
    """Sanitize input list of variable names"""
    if var_list is None:
        if arg_name == "variable_names":
            return None
        else:
            return []

    if isinstance(var_list, str):
        return [var_list]

    if not isinstance(var_list, (list, tuple)):
        raise ValueError(f"{arg_name} must be a list of strings")

    return list(var_list)



# -*- python -*- or rather like

# Copyright (c) 2001-2002 Enthought, Inc. 2003, SciPy Developers.
# All rights reserved.
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions
# are met:
#
# 1. Redistributions of source code must retain the above copyright
#    notice, this list of conditions and the following disclaimer.
#
# 2. Redistributions in binary form must reproduce the above
#    copyright notice, this list of conditions and the following
#    disclaimer in the documentation and/or other materials provided
#    with the distribution.
#
# 3. Neither the name of the copyright holder nor the names of its
#    contributors may be used to endorse or promote products derived
#    from this software without specific prior written permission.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
# "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
# LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR
# A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT
# OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
# SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
# LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
# DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
# THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
# (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
# OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

from cpython cimport PyBytes_AS_STRING, PyBytes_FromStringAndSize, PyBytes_Size


# Function to allocate, wrap memory via Python string creation
cdef inline object pyalloc_v(Py_ssize_t n, void **pp):
    cdef object ob = PyBytes_FromStringAndSize(NULL, n)
    pp[0] = <void*> PyBytes_AS_STRING(ob)
    return ob
